unit SetupViewData;


// {$define USE_SCROLLPAGER}


interface


uses

   Winapi.Windows, Winapi.Messages,
   System.SysUtils, System.Variants, System.Classes,
   MtxComCtrls,
   Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.ToolWin, Vcl.ComCtrls,
   VCLTee.TeEngine, Vcl.ExtCtrls, VCLTee.TeeGDIPlus, VCLTee.Series, VCLTee.TeCanvas, VCLTee.TeeHtml,
   VCLTee.Chart, VCLTee.TeeThemes, VCLTee.TeeProcs, VCLTee.TeeTools, VCLTee.TeeScrollPagerTool; // ,
// StoreMdf;

const
   MAX_SIGNALS = 8;

   // Asam Mdf Data Type Codes, Manual 4.2.0, p. 87f.
   // cn_data_type, UINT8
   // integer
   CN_D_UINT_LE  = 0;  // Unsigned Integer LE Byte Order
   CN_D_UINT_BE  = 1;  // Unsigned Integer BE Byte Order
   CN_D_SINT_LE  = 2;  // Signed Integer LE Byte Order
   CN_D_SINT_BE  = 3;  // Unsigned Integer BE Byte Order
   // floating-point
   CN_D_FLOAT_LE = 4;  // Float (IEEE 754) LE Byte Order
   CN_D_FLOAT_BE = 5;  // Float (IEEE 754) BE Byte Order
   // string, NULL-terminated
   CN_D_STRING   = 6;  // String (ISO-8859-1 Latin), NULL terminated)
   CN_D_UTF8     = 7;  // String (UTF8-encoded), NULL terminated)
   CN_D_UTF16_LE = 8;  // String (UTF16-LE Byte order), NULL terminated)
   CN_D_UTF16_BE = 9;  // String (UTF16-BE Byte order), NULL terminated)
   // complex
   CN_D_BYTES    = 10; // Byte array
   CN_D_SAMPLE   = 11; // MIME sample
   CN_D_STREAM   = 12; // MIME stream
   CN_D_CODATE   = 13; // CANOpen Date
   CN_D_COTIME   = 14; // CANOpen Time
   // {Mdf4.2}constexpr auto DT_COMPLEX_LE = 15; // real part followed by imaginary part, stored as two floating - point data, both with 2, 4 or 8 Byte, LE Byte order) valid since MDF 4.2.0
   // {Mdf4.2}constexpr auto DT_COMPLEX_BE = 16;



type
   TSignal = packed record
      MdfChanIdx: UInt32;    // idx of mdf channel in mdf record (id=0:time channel;id=1:first data channel, ...)
      MdfDataGrpIdx: UInt32; // idx of data group that the channel belongs to
      Serie: TChartSeries;   // TChart object that (re)presents the data from the mdf file's channel
      // ---
      MdfOffset: UInt32;     // offset of channel value in the record of the mdf file's data group
      MdfDataType: UInt32;   // Mdf Data Type, e.g. CN_D_FLOAT_LE      example: CN_D_FLOAT_LE and {4|8} bytes
      MdfBytes: UInt32;      // Data Type byte size, e.g. 8                     = {single|double}
   end;


   TBottomAxisTimeType = (xtMeas = 0, xtDate);


   TFrameViewData = class(TFrame)
      FrameScrollPanel: TScrollBox;
      ViewChart: TChart;
      Series1: TFastLineSeries;
      RecordSlider: TMtxPositionPanel;
      MdfFileDlg: TOpenDialog;
      BtnSelectFile: TButton;
      BtnSelDefFile: TButton;
      BtnStart: TButton;
      BtnPrev: TButton;
      BtnNext: TButton;
      BtnEnd: TButton;
      BtnReadData: TButton;
      EdStartRecord: TEdit;
      EdNumberRecords: TEdit;
      Label1: TLabel;
      Label2: TLabel;
      Label3: TLabel;
      Label4: TLabel;
      EdAxisMax: TEdit;
      EdAxisMin: TEdit;
      BtnUnDoZoom: TButton;
      ChkCursor: TCheckBox;
      TopPanel: TPanel;
      BottomPanel: TPanel;
      GrpBoxRecordInput: TGroupBox;
      GrpBoxDataNavigation: TGroupBox;
      GrpBoxScaleMinMax: TGroupBox;
      ChkMeasDate: TCheckBox;
      ChkReadOnSlide: TCheckBox;
      procedure BtnSelDefFileClick(Sender: TObject);
      procedure BtnSelectFileClick(Sender: TObject);
      procedure BtnUnDoZoomClick(Sender: TObject);
      procedure ChkCursorClick(Sender: TObject);
      procedure InitFile;
      procedure SeriesXValueOrDate;
      procedure CursorView_UpdateValues(X: Integer);
      procedure BtnStartClick(Sender: TObject);
      procedure BtnPrevClick(Sender: TObject);
      procedure BtnNextClick(Sender: TObject);
      procedure BtnEndClick(Sender: TObject);
      procedure BtnReadDataClick(Sender: TObject);
      procedure OnChangePosition(Sender: TObject);
      procedure OnSetPosition(Sender: TObject);
      procedure OnEditExit(Sender: TObject);
      procedure FormKeyPress(Sender: TObject; var Key: Char);
      procedure OnChartAllowScroll(Sender: TChartAxis; var AMin, AMax: TAxisValue; var AllowScroll: Boolean);
      procedure OnChartClickSeries(Sender: TCustomChart; ASerie: TChartSeries; ValueIndex: Integer; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
      procedure OnChartClickAxis(Sender: TCustomChart; Axis: TChartAxis; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
      procedure OnClickLegend(Sender: TCustomChart; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
      // procedure OnCursorToolSnapChange(Sender: TCursorTool; X, Y: Integer; const XValue, YValue: Double; Series: TChartSeries; ValueIndex: Integer);
      procedure OnCursorToolChange(Sender: TCursorTool; X, Y: Integer; const XValue, YValue: Double; Series: TChartSeries; ValueIndex: Integer);
      procedure OnZoom(Sender: TObject);
      procedure OnUndoZoom(Sender: TObject);
      procedure OnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
      procedure OnMouseUp(Sender: TObject; Shift: TShiftState; X, Y: Integer);
      procedure OnBeforeDrawAxes(Sender: TObject);
      procedure OnBeforeDrawAxis(Sender: TObject);
      procedure OnBeforeDrawSeries(Sender: TObject);
      procedure OnAfterDrawSeries(Sender: TObject);
      procedure OnBeforeDrawChart(Sender: TObject);
      procedure ChkMeasDateClick(Sender: TObject);
      procedure ChkReadOnSlideClick(Sender: TObject);
      procedure OnChartAfterDraw(Sender: TObject);
      procedure OnAxisDrawLabel(Sender: TChartAxis; var X, Y, Z: Integer; var Text: string; var DrawLabel: Boolean);
      procedure OnBottomAxisDrawLabel(Sender: TChartAxis; var X, Y, Z: Integer; var Text: string; var DrawLabel: Boolean);
      procedure OnGetAxisLabel(Sender: TChartAxis; Series: TChartSeries; ValueIndex: Integer; var LabelText: string);
      procedure OnGetNextAxisLabel(Sender: TChartAxis; LabelIndex: Integer; var LabelValue: Double; var Stop: Boolean);

   private
      // procedure BufToSeries_Single(pChanGrp: PMdfCfgChannelGroup);
      // procedure BufToSeries_Double(pChanGrp: PMdfCfgChannelGroup);
      // procedure BufToSeries_Byte(pChanGrp: PMdfCfgChannelGroup);
      // procedure BufToSeries_SInteger(pChanGrp: PMdfCfgChannelGroup);
      // procedure BufToSeries(pChanGrp: PMdfCfgChannelGroup);
   public
      procedure InitPanel;
      procedure UpdatePanel;
      procedure UpdateDpiChangeWorkarounds;
      procedure FreePanel;

      procedure AppendChannel; // wip, add a file's channel to the chart
      procedure RemoveChannel; // wip, remove a channel from the chart
   end;


implementation


uses System.Math, DateUtils;


{$R *.dfm}

var
   Filename: string;
   FilenameDefault: string;

   StartRecord, NumberRecords: Cardinal;
   MdfRecordSize, MdfNumberRecords: Cardinal;
   MdfSampleDistance: Double;
   MdfFileStartTime: TDateTime;
   Signals: TArray<TSignal>;
   NumSignals: Cardinal;
   // Scales: TArray<TChartAxis>;
   NumScales: Cardinal;
   IdxSelectedScale: Integer;
   ScrollTools: TArray<TAxisScrollTool>;
   NumScrollTools: Cardinal;

   ByteBuffer: TArray<Byte>;
   BufferLength: UInt64;

   ScrollMinX, ScrollMaxX: Double;
   SliderSpan: Double;
   ScaleWidth: Integer;
   ScalePosFactor, ScalePosOffset: Double;

   CursorTool: TCursorTool;
   CursorOldX, CursorOldY: Integer;
   RectTool: TRectangleTool;
   AnnoTool: TAnnotationTool;
{$IFDEF USE_SCROLLPAGER}
   ScrollPagerTool: TScrollPagerTool;
{$ENDIF}
   // Options
   BottomAxisTimeType: TBottomAxisTimeType;
   ReadOnSlide: Boolean;
   GridColor: TColor;
   BackgroundColor: TColor;
   ScaleColor: TColor;
   ScaleTextColor: TColor;


function SetBufferLength(NewLength: UInt64): Boolean;
begin
   if NewLength = BufferLength then
      Exit(True);
   try
      SetLength(ByteBuffer, NewLength);
   except
      on E: EOutOfMemory do
      begin
         ByteBuffer := nil;
         BufferLength := 0;
         Exit(False);
      end;
   end;
   BufferLength := NewLength;
   Result := True;
end;


function ColorToHtmlColor(AColor: TColor): string;
begin
   Result := IntToHex(ColorToRgb(AColor), 6);
   Result := '#' + Copy(Result, 5, 2) + Copy(Result, 3, 2) + Copy(Result, 1, 2);
end;


// does not work, no color change
// function HtmlColorToColor(AHtmlColor: string): TColor;
// begin
// Delete(AHtmlColor, 1, 1);
// Result := StrToInt('$' + Copy(AHtmlColor, 5, 2) + Copy(AHtmlColor, 3, 2) + Copy(AHtmlColor, 1, 2));
// end;


function HtmlColorDecorator(Text: string; Color: TColor): string;
begin
   Result := '<FONT color=' + ColorToHtmlColor(Color) + '>' + Text + '</Font>';
end;


// procedure TFrameViewData.BufToSeries_Single(pChanGrp: PMdfCfgChannelGroup);
// var
// pValue: ^Single;
// pSerie: ^TChartSeries;
// begin
// case BottomAxisTimeType of
// xtMeas:
// begin
// var
// TimeStamp: Double;
// for var SignalIdx := 0 to NumSignals - 1 do
// begin
// pSerie := @Signals[SignalIdx].Serie;
// pSerie.BeginUpdate;
// pSerie.Clear;
//
// pValue := @ByteBuffer[Signals[SignalIdx].MdfOffset];
// for var SampleIdx := 0 to NumberRecords - 1 do
// begin
// TimeStamp := (StartRecord + SampleIdx) * pChanGrp.SampleDistance;
// pSerie.AddXY(TimeStamp, pValue^);
// Inc(pValue, pChanGrp.CNNumber - 1);
// end;
// pSerie.EndUpdate;
// pSerie.ParentChart := ViewChart; // nil: remove from chart
// end;
// end;
// xtDate:
// begin
// var
// Date: TDateTime;
// for var SignalIdx := 0 to NumSignals - 1 do
// begin
// pSerie := @Signals[SignalIdx].Serie;
// pSerie.BeginUpdate;
// pSerie.Clear;
//
// pValue := @ByteBuffer[Signals[SignalIdx].MdfOffset];
// for var SampleIdx := 0 to NumberRecords - 1 do
// begin
// Date := IncMilliSecond(MdfFileStartTime, round((StartRecord + SampleIdx) * pChanGrp.SampleDistance * 1000));
// pSerie.AddXY(Date, pValue^);
// Inc(pValue, pChanGrp.CNNumber - 1);
// end;
// pSerie.EndUpdate;
// pSerie.ParentChart := ViewChart;
// end;
// end;
// end;
// end;
//
//
// procedure TFrameViewData.BufToSeries_Double(pChanGrp: PMdfCfgChannelGroup);
// var
// pValue: ^Double;
// pSerie: ^TChartSeries;
// begin
// case BottomAxisTimeType of
// xtMeas:
// begin
// var
// TimeStamp: Double;
// for var SignalIdx := 0 to NumSignals - 1 do
// begin
// pSerie := @Signals[SignalIdx].Serie;
// pSerie.BeginUpdate;
// pSerie.Clear;
//
// pValue := @ByteBuffer[Signals[SignalIdx].MdfOffset];
// for var SampleIdx := 0 to NumberRecords - 1 do
// begin
// TimeStamp := (StartRecord + SampleIdx) * pChanGrp.SampleDistance;
// pSerie.AddXY(TimeStamp, pValue^);
// Inc(pValue, pChanGrp.CNNumber - 1);
// end;
// pSerie.EndUpdate;
// pSerie.ParentChart := ViewChart; // nil: remove from chart
// end;
// end;
// xtDate:
// begin
// var
// Date: TDateTime;
// for var SignalIdx := 0 to NumSignals - 1 do
// begin
// pSerie := @Signals[SignalIdx].Serie;
// pSerie.BeginUpdate;
// pSerie.Clear;
//
// pValue := @ByteBuffer[Signals[SignalIdx].MdfOffset];
// for var SampleIdx := 0 to NumberRecords - 1 do
// begin
// Date := IncMilliSecond(MdfFileStartTime, round((StartRecord + SampleIdx) * pChanGrp.SampleDistance * 1000));
// pSerie.AddXY(Date, pValue^);
// Inc(pValue, pChanGrp.CNNumber - 1);
// end;
// pSerie.EndUpdate;
// pSerie.ParentChart := ViewChart;
// end;
// end;
// end;
// end;
//
//
// procedure TFrameViewData.BufToSeries_Byte(pChanGrp: PMdfCfgChannelGroup);
// var
// pValue: ^Byte;
// pSerie: ^TChartSeries;
// begin
// case BottomAxisTimeType of
// xtMeas:
// begin
// var
// TimeStamp: Double;
// for var SignalIdx := 0 to NumSignals - 1 do
// begin
// pSerie := @Signals[SignalIdx].Serie;
// pSerie.BeginUpdate;
// pSerie.Clear;
//
// pValue := @ByteBuffer[Signals[SignalIdx].MdfOffset];
// for var SampleIdx := 0 to NumberRecords - 1 do
// begin
// TimeStamp := (StartRecord + SampleIdx) * pChanGrp.SampleDistance;
// pSerie.AddXY(TimeStamp, Double(pValue^));
// Inc(pValue, pChanGrp.CNNumber - 1);
// end;
// pSerie.EndUpdate;
// pSerie.ParentChart := ViewChart; // nil: remove from chart
// end;
// end;
// xtDate:
// begin
// var
// Date: TDateTime;
// for var SignalIdx := 0 to NumSignals - 1 do
// begin
// pSerie := @Signals[SignalIdx].Serie;
// pSerie.BeginUpdate;
// pSerie.Clear;
//
// pValue := @ByteBuffer[Signals[SignalIdx].MdfOffset];
// for var SampleIdx := 0 to NumberRecords - 1 do
// begin
// Date := IncMilliSecond(MdfFileStartTime, round((StartRecord + SampleIdx) * pChanGrp.SampleDistance * 1000));
// pSerie.AddXY(Date, Double(pValue^));
// Inc(pValue, pChanGrp.CNNumber - 1);
// end;
// pSerie.EndUpdate;
// pSerie.ParentChart := ViewChart;
// end;
// end;
// end;
//
// end;
//
//
// procedure TFrameViewData.BufToSeries_SInteger(pChanGrp: PMdfCfgChannelGroup);
// var
// pValue: ^Int32;
// pSerie: ^TChartSeries;
// begin
// case BottomAxisTimeType of
// xtMeas:
// begin
// var
// TimeStamp: Double;
// for var SignalIdx := 0 to NumSignals - 1 do
// begin
// pSerie := @Signals[SignalIdx].Serie;
// pSerie.BeginUpdate;
// pSerie.Clear;
//
// pValue := @ByteBuffer[Signals[SignalIdx].MdfOffset];
// for var SampleIdx := 0 to NumberRecords - 1 do
// begin
// TimeStamp := (StartRecord + SampleIdx) * pChanGrp.SampleDistance;
// pSerie.AddXY(TimeStamp, Double(pValue^));
// Inc(pValue, pChanGrp.CNNumber - 1);
// end;
// pSerie.EndUpdate;
// pSerie.ParentChart := ViewChart; // nil: remove from chart
// end;
// end;
// xtDate:
// begin
// var
// Date: TDateTime;
// for var SignalIdx := 0 to NumSignals - 1 do
// begin
// pSerie := @Signals[SignalIdx].Serie;
// pSerie.BeginUpdate;
// pSerie.Clear;
//
// pValue := @ByteBuffer[Signals[SignalIdx].MdfOffset];
// for var SampleIdx := 0 to NumberRecords - 1 do
// begin
// Date := IncMilliSecond(MdfFileStartTime, round((StartRecord + SampleIdx) * pChanGrp.SampleDistance * 1000));
// pSerie.AddXY(Date, Double(pValue^));
// Inc(pValue, pChanGrp.CNNumber - 1);
// end;
// pSerie.EndUpdate;
// pSerie.ParentChart := ViewChart;
// end;
// end;
// end;
// end;
//
//
// procedure TFrameViewData.BufToSeries(pChanGrp: PMdfCfgChannelGroup);
// var
// pByte: ^Byte;
// pValue: ^Double;
// pSerie: ^TChartSeries;
// begin
// case BottomAxisTimeType of
// xtMeas:
// begin
// var
// TimeStamp: Double;
// for var SignalIdx := 0 to NumSignals - 1 do
// begin
// pSerie := @Signals[SignalIdx].Serie;
// pSerie.BeginUpdate;
// pSerie.Clear;
//
// pValue := @ByteBuffer[Signals[SignalIdx].MdfOffset];
// for var SampleIdx := 0 to NumberRecords - 1 do
// begin
// TimeStamp := (StartRecord + SampleIdx) * pChanGrp.SampleDistance;
// pSerie.AddXY(TimeStamp, pValue^);
// Inc(pValue, pChanGrp.CNNumber - 1);
// end;
// pSerie.EndUpdate;
// pSerie.ParentChart := ViewChart; // nil: remove from chart
// end;
// end;
// xtDate:
// begin
// var
// Date: TDateTime;
// for var SignalIdx := 0 to NumSignals - 1 do
// begin
// pSerie := @Signals[SignalIdx].Serie;
// pSerie.BeginUpdate;
// pSerie.Clear;
//
// pValue := @ByteBuffer[Signals[SignalIdx].MdfOffset];
// for var SampleIdx := 0 to NumberRecords - 1 do
// begin
// Date := IncMilliSecond(MdfFileStartTime, round((StartRecord + SampleIdx) * pChanGrp.SampleDistance * 1000));
// pSerie.AddXY(Date, pValue^);
// Inc(pValue, pChanGrp.CNNumber - 1);
// end;
// pSerie.EndUpdate;
// pSerie.ParentChart := ViewChart;
// end;
// end;
// end;
// end;


procedure TFrameViewData.InitPanel;
begin
   Signals := nil;
   NumSignals := 0;
   ByteBuffer := nil;
   BufferLength := 0;
   ReadOnSlide := False;

   BottomAxisTimeType := xtMeas; //xtDate;
   // IdxGridScale := -1;
   IdxSelectedScale := -1;
   ScaleWidth := -1;
   ScalePosFactor := 4.0;
   ScalePosOffset := 1.0;

   FrameScrollPanel.Left := 0;
   FrameScrollPanel.Top := 0;
   FrameScrollPanel.Width := Self.Width;
   FrameScrollPanel.Height := Self.Height;
   FrameScrollPanel.Anchors := [akLeft, akTop, akRight, akBottom];
   FrameScrollPanel.AutoScroll := True;
   // FrameScrollPanel.Font.Name := DmSettings.UiFontName;
   // FrameScrollPanel.Font.Height := -MulDiv(DmSettings.UiFontSize, Self.CurrentPPI, 72);

   ViewChart.Left := 0;
   ViewChart.Top := 0;
   ViewChart.Width := FrameScrollPanel.Width;
   ViewChart.Height := FrameScrollPanel.Height;
   ViewChart.Anchors := [akLeft, akTop, akRight, akBottom];
   ViewChart.Walls.Show;
   ViewChart.Border.Show;
   ViewChart.View3DWalls := False;
   ViewChart.BevelInner := bvNone;
   ViewChart.BevelOuter := bvNone;
   ViewChart.View3D := False;
   ViewChart.RemoveAllSeries;        // remove default "Series1" from chart (tchart crashs on init if there is no default series)
   ViewChart.Zoom.Animated := False;
   ViewChart.AllowPanning := pmBoth; // pmVertical;
   ViewChart.Title.Hide;
   ViewChart.ClipPoints := True;
   ViewChart.Color := BackgroundColor;
   ViewChart.MarginLeft := 3;   // 3=default "These properties are expressed in screen pixels." -- nee, wohl eher nicht...
{$IFDEF USE_SCROLLPAGER}
   ViewChart.MarginTop := 15;   // add space between chart and its scrollpager (less tick text overlapping)
   ViewChart.MarginBottom := 1; // remove empty space between bottom axis and the control panel by setting a value (no automatic)
{$ENDIF}
   ViewChart.Legend.Font.Quality := fqClearType;
   // ViewChart.Legend.Font.Name := DmSettings.UiFontName;
   // ViewChart.Legend.Font.SizeFloat := -MulDiv(DmSettings.UiFontSize, CurrentPPI, 72);
   ViewChart.Legend.Color := BackgroundColor; //
   ViewChart.Legend.Frame.Color := ScaleColor;

   // do not show chart's default axes: left, right, top     (using 'custom axes' for the left side)
   ViewChart.LeftAxis.Visible := False;
   ViewChart.RightAxis.Visible := False;
   ViewChart.TopAxis.Visible := False;

   ViewChart.LeftAxis.Grid.Visible := False; // grid will be attached to custom axis

   // x
   ViewChart.BottomAxis.Visible := False;    // shown again if after file was loaded!
   ViewChart.BottomAxis.Grid.Color := GridColor;
   ViewChart.BottomAxis.LabelsOnAxis := True;
   ViewChart.BottomAxis.MinorTickCount := 1;
   ViewChart.BottomAxis.Title.Text := '';
   ViewChart.BottomAxis.Title.Position := tpEnd;
   ViewChart.BottomAxis.Title.Alignment := taRightJustify;
   // ViewChart.BottomAxis.LabelsFont.Name := DmSettings.UiFontName;
   // ViewChart.BottomAxis.LabelsFont.SizeFloat := -MulDiv(DmSettings.ScaleFontSize, CurrentPPI, 72);
   ViewChart.BottomAxis.LabelsFont.Quality := fqClearType;
   ViewChart.BottomAxis.LabelsFont.Color := ScaleTextColor;
   ViewChart.BottomAxis.Axis.Color := ScaleColor;
   ViewChart.BottomAxis.Ticks.Color := ScaleColor;

   ViewChart.BottomAxis.OnDrawLabel := OnBottomAxisDrawLabel; // test
   ViewChart.OnGetAxisLabel := OnGetAxisLabel;                // test
   ViewChart.OnGetNextAxisLabel := OnGetNextAxisLabel;        // test
   // ViewChart.OnGetAxisLabel(


   case BottomAxisTimeType of
      xtMeas:
         begin
            ViewChart.BottomAxis.AxisValuesFormat := '#,##0.##';
         end;
      xtDate:
         begin
            ViewChart.BottomAxis.DateTimeFormat := 'hh:mm:ss,zzz' + #13 + 'ddddd';
         end;
   end;

   CursorTool := TCursorTool.Create(ViewChart);
   CursorTool.ParentChart := nil;
   CursorTool.Series := nil;
   CursorTool.Style := cssBoth; // cssScope //cssVertical;
   CursorTool.Pen.Color := ScaleColor;
   CursorTool.Active := False;
   CursorTool.Snap := True;
   CursorTool.SnapStyle := ssVertical; // ssHorizontal = ssDefault
   // CursorTool.FollowMouse := True;
   CursorTool.FullRepaint := False;    // less flickering on fast movement
   CursorTool.OnChange := OnCursorToolChange;

   // displays cursor tool's information
   RectTool := TRectangleTool.Create(ViewChart);
   RectTool.ParentChart := nil;
   RectTool.Visible := True;
   RectTool.Active := False;
   RectTool.AutoSize := True;
   RectTool.Position := ppRightTop;
   RectTool.Text := '';
   RectTool.TextAlignment := taRightJustify;
   RectTool.AllowResize := False;
   RectTool.Draw3D := False;
   RectTool.Selected.Hover.Visible := False; // removes red 'selection' border on drag
   RectTool.Shape.Color := BackgroundColor;
   RectTool.Shape.ClipText := True;
   RectTool.Shape.Transparency := 0;
   RectTool.Shape.TextFormat := ttfHtml;
   RectTool.Shape.Shadow.Show;
   RectTool.Shape.Shadow.Size := 4;
   RectTool.Shape.Font.Name := 'Consolas';
   // RectTool.Shape.Font.Size := MulDiv(DmSettings.UiFontSize, Self.CurrentPPI, 72);
   RectTool.Shape.Font.Quality := fqClearType;
   RectTool.Shape.Font.Color := ScaleColor; // color of X value in cursor value box, HTMLColor does not work for some reason
   RectTool.Shape.Frame.Color := ScaleColor;

   // text hint 'select file'
   AnnoTool := TAnnotationTool.Create(ViewChart);
   AnnoTool.ParentChart := ViewChart;
   AnnoTool.Visible := True;
   AnnoTool.Active := True;
   AnnoTool.AutoSize := True;
   AnnoTool.Position := ppCenter;
   AnnoTool.Text := 'Please select a file!'; // rsSetupViewData_SelectFileHint;
   AnnoTool.TextAlignment := taCenter;
   AnnoTool.Draw3D := False;
   AnnoTool.Shape.Shadow.Show;
   AnnoTool.Shape.Shadow.Size := 4;
   // AnnoTool.Shape.Font.Name := DmSettings.UiFontName;
   // AnnoTool.Shape.Font.SizeFloat := -MulDiv(DmSettings.UiFontSize, CurrentPPI, 72);
   AnnoTool.Shape.Font.Quality := fqClearType;
   AnnoTool.Shape.Frame.Color := ScaleColor;

{$IFDEF USE_SCROLLPAGER}
   // test scrollpager
   ScrollPagerTool := TScrollPagerTool.Create(Self);
   ScrollPagerTool.ParentChart := ViewChart;
   ScrollPagerTool.Visible := False;     // hidden until a file is selected
   ScrollPagerTool.DivisionRatio := 8.0; // ScrollPager is part of the chart and consumes a bit of its space
   ScrollPagerTool.SubChartTChart.BottomAxis.Grid.Hide;
   // ScrollPagerTool.SubChartTChart.LeftAxis.Grid.Hide;
   ScrollPagerTool.SubChartTChart.Frame.Visible := False;
   ScrollPagerTool.SubChartTChart.View3D := False;
   ScrollPagerTool.ColorBandTool.StartLine.Active := True;
   ScrollPagerTool.ColorBandTool.EndLine.Active := True;
   ScrollPagerTool.StartLinePointer.Size := 6; // *dpi?
   ScrollPagerTool.EndLinePointer.Size := 6;
   ScrollPagerTool.Align := saBefore;          // show ScrollPager above (=before) the chart, default: below
   ScrollPagerTool.SubChartTChart.BottomAxis.LabelsFont.Color := ScaleColor;
{$ENDIF}
   BtnSelDefFile.Hint := FilenameDefault;
   BtnSelDefFile.ShowHint := True;
   BtnReadData.Hint := 'Read Data';
   BtnReadData.ShowHint := True;

   TopPanel.Color := BackgroundColor;
   RecordSlider.Caption := '';
   RecordSlider.Enabled := False;
   RecordSlider.Color := BackgroundColor;
   RecordSlider.SliderColor := ScaleColor;

   BtnPrev.Enabled := False;
   BtnNext.Enabled := False;
   BtnStart.Enabled := False;
   BtnEnd.Enabled := False;
   BtnReadData.Enabled := False;
   BtnUnDoZoom.Enabled := False;
   ChkCursor.Enabled := False;
   ChkMeasDate.Enabled := False;
   ChkReadOnSlide.Enabled := False;
   EdStartRecord.Enabled := False;
   EdNumberRecords.Enabled := False;
   EdAxisMax.Enabled := False;
   EdAxisMin.Enabled := False;

   ChkCursor.Checked := False;
   ChkReadOnSlide.Checked := ReadOnSlide;
   ChkMeasDate.Checked := (BottomAxisTimeType = xtDate);
end;


procedure TFrameViewData.UpdatePanel;
begin
end;


procedure TFrameViewData.UpdateDpiChangeWorkarounds;
begin
end;


procedure TFrameViewData.FreePanel;
begin
end;


procedure TFrameViewData.AppendChannel;
begin

end;


procedure TFrameViewData.RemoveChannel;
begin

end;


// todo, error messages (no Memo control anymore)
procedure TFrameViewData.InitFile;
var
   // pDataGrp: PMdfCfgDataGroup;
   // pChanGrp: PMdfCfgChannelGroup;
   MdfNumSignalsScalar: UInt32; // MdfNumSignals, MdfNumSignalsCAN
   MdfNumDgScalar: UInt32;      // MdfNumDgCan
   DgIdx, DGDataType: UInt32;
   SignalIdx: UInt32;
   ValueFormat: string;
   Delta: Double;
   Decimals: Integer;
begin
   if Filename = '' then
   begin
      // errmsg
      Exit;
   end;
   // if not MdfReadInit(Filename) then
   // begin
   // // errmsg
   // Exit;
   // end;

   // fixme, populate some control with file details, show on-demand, button/menu etc
   // PrintMdfFileInfo(Memo1.Lines);

   // dfDecodeTime(pMdfMeas.FileStartTime[0], MdfFileStartTime);

   // prepare existing chart for new file
   ViewChart.RemoveAllSeries; // does not delete series objects!
   if NumSignals > 0 then
   begin
      for var Idx := 0 to NumSignals - 1 do
         if Signals[Idx].Serie <> nil then
            Signals[Idx].Serie.Delete(Idx); // Signals[Idx].Serie.Free;
      NumSignals := 0;
   end;

   ViewChart.BottomAxis.Visible := True;


   // count channels of data groups with scalar data type (single, double, ...)
   DgIdx := 0;
   MdfNumSignalsScalar := 0;
   MdfNumDgScalar := 0;
   // while DgIdx < pMdfMeas.DGNumber do
   // begin
   // pDataGrp := @(pMdfMeas.DataGroups[DgIdx]);
   // pChanGrp := @(pDataGrp.ChannelGroups[0]);    // demeas creates sorted Mdf files: single CG only
   //
   // DGDataType := pChanGrp.Channels[1].DataType; // first data channel (0: time channel)
   // case DGDataType of
   // CN_D_FLOAT_LE, CN_D_UINT_LE, CN_D_SINT_LE:
   // begin
   // Inc(MdfNumDgScalar);
   // Inc(MdfNumSignalsScalar, pChanGrp.CNNumber - 1);
   // end;
   // CN_D_BYTES:
   // break; // CAN data groups approching
   // else
   // break;
   // end;
   // Inc(DgIdx);
   // end;

   MdfNumSignalsScalar := 2;
   MdfNumDgScalar := 1;

   // fixme, handle data groups with CAN data

   if MdfNumSignalsScalar = 0 then
      Exit;

   NumSignals := MdfNumSignalsScalar; // MdfNumSignals = MdfNumSignalsScalar + MdfNumSignalsCAN
   SetLength(Signals, NumSignals);

   // fixme, handle mixed data groups, e.g. several data groups with different sample rate and CAN data

   SignalIdx := 0;
   for DgIdx := 0 to MdfNumDgScalar - 1 do
   begin
      // pDataGrp := @(pMdfMeas.DataGroups[DgIdx]);
      // pChanGrp := @(pDataGrp.ChannelGroups[0]);
      for var CnIdx := 1 to 2 do                          // pChanGrp.CNNumber - 1 do                          // skip virtual time channel
      begin
         Signals[SignalIdx].MdfChanIdx := CnIdx;          // where do I come from?
         Signals[SignalIdx].MdfDataGrpIdx := DgIdx;       // "
         Signals[SignalIdx].MdfOffset := (CnIdx - 1) * 4; // pChanGrp.Channels[CnIdx].ByteOffset;
         Signals[SignalIdx].MdfBytes := 4;                // pChanGrp.Channels[CnIdx].ByteSize;
         Signals[SignalIdx].MdfDataType := CN_D_FLOAT_LE; // pChanGrp.Channels[CnIdx].DataType; // Mdf data type, e.g. CN_D_FLOAT_LE, not the Demeas one
         Signals[SignalIdx].Serie := TFastLineSeries.Create(ViewChart);
         Signals[SignalIdx].Serie.Title := 'A-Ai' + (CnIdx - 1).ToString; // PWChar(pChanGrp.Channels[CnIdx].ChanName);
         Signals[SignalIdx].Serie.Pen.Width := 1;
         Signals[SignalIdx].Serie.BeforeDrawValues := OnBeforeDrawSeries;
         Signals[SignalIdx].Serie.AfterDrawValues := OnAfterDrawSeries;

         TFastLineSeries(Signals[SignalIdx].Serie).FastPen := True;
         TFastLineSeries(Signals[SignalIdx].Serie).DrawAllPoints := False; // do not draw multiple values to the same x-position
         TFastLineSeries(Signals[SignalIdx].Serie).ClickTolerance := 2;

         case BottomAxisTimeType of
            xtMeas:
               Signals[SignalIdx].Serie.XValues.DateTime := False;
            xtDate:
               Signals[SignalIdx].Serie.XValues.DateTime := True;
         end;
         Inc(SignalIdx);
      end;
   end;

   // fixme, handle data groups with CAN data

   if ViewChart.CustomAxes.Count > 0 then
   begin
      while ViewChart.CustomAxes.Count > 0 do
         ViewChart.CustomAxes.Delete(0);
      NumScales := 0; // =axis
   end;

   NumScales := NumSignals;
   // SetLength(Scales, NumScales);
   // SetLength(Scales, NumScales + 1); // additional hidden 'grid scale' at the end, does not work
   NumScrollTools := NumSignals;
   SetLength(ScrollTools, NumScrollTools);

   for var Idx := 0 to NumScales - 1 do
   begin
      ViewChart.CustomAxes.Add();

      with ViewChart.CustomAxes[Idx] do
      begin
         // does not work
         // Scales[Idx].Texts.MarginToAxis := -170; // 100; -70;
         // ViewChart.CustomAxes[Idx].Texts.MarginToAxis := -70;
         Grid.Hide;
         Grid.Color := clWhite; //GridColor; // test, clWhite
         MinorTickCount := 1;
         SetMinMax(-10.0, 10.0);
         Increment := 2.5;
         Title.Text := '';
         Title.Position := tpCenter;
         Title.Alignment := taRightJustify;
         // Title.Font.Name := DmSettings.UiFontName;
         // Title.Font.SizeFloat := -MulDiv(DmSettings.ScaleFontSize, CurrentPPI, 72);
         PositionPercent := Idx * ScalePosFactor + ScalePosOffset; // vor muldiv fontsize: 3% der chartbreite fr eine achse
         MinimumOffset := 40;
         MaximumOffset := 40;
         // LabelsFont.Name := DmSettings.UiFontName;
         // LabelsFont.SizeFloat := -MulDiv(DmSettings.ScaleFontSize, CurrentPPI, 72);
         LabelsFont.Quality := fqClearType;
         LabelsAngle := 90;
         LabelsFont.Color := ScaleTextColor;

         Axis.Color := ScaleColor;
         Ticks.Color := ScaleColor;

         OnDrawLabel := OnAxisDrawLabel;

         // link grid to first scale (series/axis)
         if Idx = 0 then
         begin
            Grid.Show;
            Grid.Width := MulDiv(1, CurrentPPI, 72); // 1
         end;

         Delta := Maximum - Minimum;
         ValueFormat := '#,##0.';
         Decimals := 1;
         while ((Delta / Items.Count) * (Decimals * 10) < 1) and (Decimals < 13) do
         begin
            Inc(Decimals);
            ValueFormat := ValueFormat + '#';
         end;
         // 0 at pos x: show at least this amount of decimals, trailing zeros
         // '1.1' --> xyz.##0 --> '1.100'
         ValueFormat := ValueFormat + '0';
         AxisValuesFormat := ValueFormat;

         ScrollTools[Idx] := TAxisScrollTool.Create(ViewChart);
         ScrollTools[Idx].Axis := ViewChart.CustomAxes[Idx];

         // attach scale to series
         Signals[Idx].Serie.CustomVertAxis := ViewChart.CustomAxes[Idx];
         Title.Text := Signals[Idx].Serie.Title;
      end;
   end;



   ViewChart.BottomAxis.StartPosition := ScalePosFactor * (NumScales - 1) + ScalePosOffset;
   ViewChart.BottomAxis.EndPosition := 100.0;

   // fixme
   // - handle multiple data groups (with identical number of samples only!)
   // - handle data groups with different sample rates
   DgIdx := 0;
   // pDataGrp := @(pMdfMeas.DataGroups[DgIdx]);
   // pChanGrp := @(pDataGrp.ChannelGroups[0]);

   MdfSampleDistance := 0.000001; // pChanGrp.SampleDistance;
   MdfNumberRecords := 1000;      // pChanGrp.RecordNumber;
   MdfRecordSize := 8;            // pChanGrp.RecordSize;
   StartRecord := 0;
   NumberRecords := 1000;
   // if pChanGrp.RecordNumber < 100000 then
   // NumberRecords := MdfNumberRecords;
   SetBufferLength(MdfRecordSize * NumberRecords);

   EdStartRecord.Text := StartRecord.ToString;
   EdNumberRecords.Text := NumberRecords.ToString;

   // fixme
   // handle different record numbers for different data groups,
   // but there is only one slider...!?
   RecordSlider.SliderMin := 0.0;
   RecordSlider.SliderMax := MdfNumberRecords;
   RecordSlider.SliderPosition := StartRecord;
   RecordSlider.SliderSpan := NumberRecords;
   SliderSpan := NumberRecords;

   RecordSlider.Enabled := True;
   EdStartRecord.Enabled := True;
   EdNumberRecords.Enabled := True;
   BtnStart.Enabled := True;
   BtnEnd.Enabled := True;
   BtnPrev.Enabled := True;
   BtnNext.Enabled := True;
   BtnReadData.Enabled := True;
   BtnUnDoZoom.Enabled := True;
   ChkCursor.Enabled := True;
   ChkMeasDate.Enabled := True;
   ChkReadOnSlide.Enabled := True;

   AnnoTool.Active := False;

   Self.BtnReadDataClick(Self);
   ViewChart.Invalidate;


{$IFDEF USE_SCROLLPAGER}
   // Funktioniert immer noch nicht:
   // - in der Vorschau/ScrollPager werden keine Signale dargestellt
   // - kein Y-Achse im ScrollPager
   // - Anfang/Ende des Griffs vom Scrollpager programmatisch auf Startwerte setzen
   // >> ???

   ScrollPagerTool.Visible := False;

   ScrollPagerTool.SubChartTChart.RemoveAllSeries;

   // select series from main chart to be shown as preview in ScrollPager's sub chart
   // ScrollPagerTool.Series := ViewChart[0];
   ScrollPagerTool.AddSeries(ViewChart[0]);
   // ScrollPagerTool.Series.Visible := True;

   // ScrollPagerTool.SubChartTChart.LeftAxis.SetMinMax(-10.0, 10.0);
   // ScrollPagerTool.SubChartTChart.LeftAxis.Automatic := True;

   // set left axis of series of main chart to be used as left axis of series in Scrollpager's sub chart
   // ScrollPagerTool.Series.VertAxis := ViewChart[0].VertAxis;
   ScrollPagerTool.SubChartTChart.Axes.Left.Assign(Signals[0].Serie.GetVertAxis);
   ScrollPagerTool.SubChartTChart.LeftAxis.Visible := True;

   ScrollPagerTool.SubChartTChart.BottomAxis.DateTimeFormat := ViewChart.BottomAxis.DateTimeFormat;
   ScrollPagerTool.SubChartTChart.BottomAxis.AxisValuesFormat := ViewChart.BottomAxis.AxisValuesFormat;
   ScrollPagerTool.SubChartTChart.BottomAxis.Automatic := False;
   ScrollPagerTool.SubChartTChart.BottomAxis.AutomaticMinimum := True;
   ScrollPagerTool.SubChartTChart.BottomAxis.AutomaticMaximum := True;

   // show labels with numeric value of start/end position's Y value of series
   ScrollPagerTool.ColorBandTool.StartLine.AnnotationValue := True;
   ScrollPagerTool.ColorBandTool.EndLine.AnnotationValue := True;

   // determine beginning/ending of Scrollpager's handle
   var
      XStart, XEnd: Double;
   XStart := ViewChart[0].XValue[0];                                 // first x value of first series in main chart
   XEnd := ViewChart[0].XValue[(Signals[0].Serie.Count - 1) div 10]; // last value/record

   ScrollPagerTool.ColorBandTool.StartValue := XStart;
   ScrollPagerTool.ColorBandTool.EndValue := XEnd;

   ScrollPagerTool.SubChartTChart.BottomAxis.Increment := (XEnd - XStart) / 5.0;

   // ScrollPagerTool.SubChartTChart.Top := ScrollPagerTool.SubChartTChart.Top + 300;

   ScrollPagerTool.Visible := True;
   ScrollPagerTool.SetUpScrollPager;


   // var
   // c: Integer;
   // var
   // d1, d2: Double;
   // c := ScrollPagerTool.Series.Count;
   //
   // d1 := ScrollPagerTool.Series.XValue[0];
   // d2 := ViewChart[0].XValue[0];
   //
   // d1 := ScrollPagerTool.Series.YValue[5];
   // d2 := ViewChart[0].YValue[5];
   //
   // // c := ScrollPagerTool.Series.VisibleCount; // = 0, wieso?
   //
   // TFastLineSeries(ScrollPagerTool.SubChartTChart[0]).DrawStyle := flAll;
   //
   // ScrollPagerTool.SubChartTChart[0].DrawBetweenPoints := True;
   // ScrollPagerTool.SubChartTChart[0].Active := True;
   // ScrollPagerTool.Series.Visible := True;
   // ScrollPagerTool.SubChartTChart.LeftAxis.Automatic := True;
   // ScrollPagerTool.SubChartTChart.LeftAxis.Visible := True;
   // ScrollPagerTool.SubChartTChart.LeftAxis.SetMinMax(0, 10);
   //
   // Inc(c);
{$ENDIF}

   // ViewChart.Invalidate;

   // ---

   // ScrollPagerTool.ColorBandTool.StartValue := ViewChart[0].
   // ScrollPagerTool.ColorBandTool.StartValue := ViewChart[0].GetVertAxis.
   // ScrollPagerTool.ColorBandTool.EndValue := ScrollPagerTool.ColorBandTool.StartValue + 100000;

   // ScrollPagerTool.ParentChart := nil;
   // ScrollPagerTool.ParentChart := ViewChart;

   // ++changes color of series in Main Chart...?!
   // ScrollPagerTool.Series.Pen.Color := clWhite;
   // ScrollPagerTool.Series.SeriesColor := clWhite;
   // ++beschriftung des scrollpager-skala ausblenden, aber skala/ticks behalten -- geht nicht, alles oder nichts
   // ScrollPagerTool.SubChartTChart.BottomAxis.Labels := False;
   // ScrollPagerTool.SubChartTChart.BottomAxis.LabelStyle := talValue; // talMark; // talNone;
   // ScrollPagerTool.SubChartTChart.BottomAxis.Items.Clear;


   // // additional hidden 'grid scale'
   // Scales[NumScales] := ViewChart.CustomAxes.Add();
   // // Scales[NumScales] := Scales[NumScales - 1];
   // Scales[NumScales].PositionPercent := Scales[NumScales - 1].PositionPercent;
   //
   // Scales[NumScales].Minimum := Scales[0].Minimum;
   // Scales[NumScales].Maximum := Scales[0].Maximum;
   // Scales[NumScales].MinorTickCount := Scales[0].MinorTickCount;
   // Scales[NumScales].Increment := Scales[0].Increment;
   // Scales[NumScales].MinimumOffset := Scales[0].MinimumOffset;
   // Scales[NumScales].MaximumOffset := Scales[0].MaximumOffset;
   // Scales[NumScales].OnDrawLabel := OnAxisDrawLabel;
end;


procedure TFrameViewData.SeriesXValueOrDate;
var
   toDate: Boolean;
begin
   if BottomAxisTimeType = xtDate then
      toDate := True
   else
      toDate := False;
   for var Idx := 0 to NumSignals - 1 do
      Signals[Idx].Serie.XValues.DateTime := toDate;
end;


procedure TFrameViewData.CursorView_UpdateValues(X: Integer);
var
   XDate: TDateTime;
   XValue, Delta: Double;
   YValueIdx: Integer;
   S, StrName, StrUnit, StrNumber: string;
begin
   // get Y values for all signals based on cursor's X position
   // X: works only if all series have the same x-axis
   // Y: get series value index that corresponds to cursor's X position
   XValue := CursorTool.Series.XScreenToValue(X);            // use serie that is linked to cursor
   YValueIdx := CursorTool.NearestPoint(cssVertical, Delta); // delta=0 because cursor-snapping-to-Xpos is true

   S := '';
   case BottomAxisTimeType of
      xtMeas:
         begin
            StrUnit := '[s]';
            S := S + 'X: ' + FloatToStrF(XValue, ffFixed, 10, 3);
            S := S + ' ' + StrUnit;
            // ... + HtmlColorDecorator(S, ScaleColor); // wieso tut's das HIER nicht, unten aber schon? (innerhalb Schleife)
            for var Idx := 0 to NumSignals - 1 do
            begin
               if Signals[Idx].Serie.Visible then
               begin
                  S := S + #13;
                  if Signals[Idx].Serie = CursorTool.Series then
                     S := S + '*'
                  else
                     S := S + ' ';
                  StrName := Signals[Idx].Serie.Title + ': ';
                  StrNumber := FloatToStrF(Signals[Idx].Serie.YValue[YValueIdx], ffFixed, 10, 3);
                  StrNumber := Format('%10s', [StrNumber]);
                  S := S + HtmlColorDecorator(StrName + StrNumber, Signals[Idx].Serie.Color);
               end;
            end;
         end;
      xtDate:
         begin
            XDate := XValue;
            S := S + 'X: ' + XDate.Format('hh:mm:ss') + ',' + XDate.GetMilliSecond.ToString;
            // ... + HtmlColorDecorator(S, ScaleColor); // wieso tut's das HIER nicht, unten aber schon? (innerhalb Schleife)
            for var Idx := 0 to NumSignals - 1 do
            begin
               if Signals[Idx].Serie.Visible then
               begin
                  S := S + #13;
                  if Signals[Idx].Serie = CursorTool.Series then
                     S := S + '*'
                  else
                     S := S + ' ';
                  StrName := Signals[Idx].Serie.Title + ': ';
                  StrNumber := FloatToStrF(Signals[Idx].Serie.YValue[YValueIdx], ffFixed, 10, 3);
                  StrNumber := Format('%15s', [StrNumber]);
                  S := S + HtmlColorDecorator(StrName + StrNumber, Signals[Idx].Serie.Color);
               end;
            end;
         end;
   end;

   RectTool.Shape.Text := S;
end;


procedure TFrameViewData.BtnReadDataClick(Sender: TObject);
var
   // pDataGrp: PMdfCfgDataGroup;
   // pChanGrp: PMdfCfgChannelGroup;
   pSerie: ^TChartSeries;
   pScale: ^TChartAxis;
   Scale: TChartAxis;
   ReducStage, Decimals: UInt32;
   DGDataType, DGDataTypeSize: UInt32;
   BytesRead: UInt64;
   AxMin, AxMax, Delta: Double;
   ValueFormat: string;
   Success: Boolean;
   pValue: ^Single;

begin

   case BottomAxisTimeType of
      xtMeas:
         begin
            var
               TimeStamp: Double;
            for var SignalIdx := 0 to NumSignals - 1 do
            begin
               pSerie := @Signals[SignalIdx].Serie;
               pSerie.BeginUpdate;
               pSerie.Clear;

               pValue := @ByteBuffer[Signals[SignalIdx].MdfOffset];
               for var SampleIdx := 0 to NumberRecords - 1 do
               begin
                  TimeStamp := (StartRecord + SampleIdx) * 0.000001; // pChanGrp.SampleDistance;
                  //pSerie.AddXY(TimeStamp, pValue^);
                  pSerie.AddXY(TimeStamp, Double(random(16)-8));
                  Inc(pValue, 2);                                    // pChanGrp.CNNumber - 1);
               end;
               pSerie.EndUpdate;
               pSerie.ParentChart := ViewChart;                      // nil: remove from chart
            end;
         end;
      xtDate:
         begin
            var
               Date: TDateTime;
            for var SignalIdx := 0 to NumSignals - 1 do
            begin
               pSerie := @Signals[SignalIdx].Serie;
               pSerie.BeginUpdate;
               pSerie.Clear;

               pValue := @ByteBuffer[Signals[SignalIdx].MdfOffset];
               for var SampleIdx := 0 to NumberRecords - 1 do
               begin
                  Date := IncMilliSecond(MdfFileStartTime, round((StartRecord + SampleIdx) * 0.000001 * 1000)); // pChanGrp.SampleDistance;
                  //pSerie.AddXY(Date, pValue^);
                  pSerie.AddXY(Date, Double(random(16)-8));
                  Inc(pValue, 2); // pChanGrp.CNNumber - 1);
               end;
               pSerie.EndUpdate;
               pSerie.ParentChart := ViewChart;
            end;
         end;
   end;



   AxMin := 0.0;
   AxMax := 0.0;

   // y
   for var Idx := 0 to NumSignals - 1 do
   begin
      pSerie := @Signals[Idx].Serie; // 1:1 relation for now
      // pScale := @Scales[Idx];
      Scale := ViewChart.CustomAxes[Idx];
      pScale := @Scale;

      pScale.Title.Font.Color := pSerie.Color;
      if Idx = 0 then
      begin
         EdAxisMin.Text := FloatToStrF(AxMin, ffFixed, 10, 3);
         EdAxisMax.Text := FloatToStrF(AxMax, ffFixed, 10, 3);
      end;
   end;

   // x
   ViewChart.BottomAxis.Automatic := False;
   ViewChart.BottomAxis.AutomaticMinimum := False;
   ViewChart.BottomAxis.AutomaticMaximum := False;
   case BottomAxisTimeType of
      xtMeas:
         begin
            AxMin := StartRecord * 0.000001;                    // pChanGrp.SampleDistance;
            AxMax := (StartRecord + NumberRecords) * 0.000001;; // pChanGrp.SampleDistance;
            ViewChart.BottomAxis.Increment := (AxMax - AxMin) / 5.0;
            ViewChart.BottomAxis.SetMinMax(AxMin, AxMax);

            Delta := ViewChart.BottomAxis.Maximum - ViewChart.BottomAxis.Minimum;
            ValueFormat := '#,##0.#';
            Decimals := 1;
            while ((Delta / ViewChart.BottomAxis.Items.Count) * (Decimals * 10) < 1) and (Decimals < 13) do
            begin
               Inc(Decimals);
               ValueFormat := ValueFormat + '#';
            end;
            ViewChart.BottomAxis.AxisValuesFormat := ValueFormat;
         end;
      xtDate:
         begin
            // Unfortunately, TDateTime has 'only' Millisecond resolution
            AxMin := IncMilliSecond(MdfFileStartTime, round(StartRecord * 0.000001 * 1000));
            AxMax := IncMilliSecond(MdfFileStartTime, round((StartRecord + NumberRecords) * 0.000001 * 1000));
            ViewChart.BottomAxis.Increment := (AxMax - AxMin) / 5.0;
            ViewChart.BottomAxis.SetMinMax(AxMin, AxMax);

            ViewChart.BottomAxis.DateTimeFormat := 'hh:mm:ss.zzz' + #13 + 'ddddd';
         end;
   end;

   ScrollMinX := AxMin;
   ScrollMaxX := AxMax;
   EdAxisMax.Enabled := True;
   EdAxisMin.Enabled := True;

   if CursorTool.Active then
   begin
      CursorTool.XValue := 0.5 * (ViewChart.BottomAxis.Maximum - ViewChart.BottomAxis.Minimum) + ViewChart.BottomAxis.Minimum;
      CursorView_UpdateValues(ViewChart.BottomAxis.CalcPosValue(CursorTool.XValue));
   end;

   // ---
   // AddXY does NOT replace/update existing entry:  multiple y-values at x-pos?!
   // pSerie.AddXY(StartRecord + SampleIdx, pValue^); // todo, x-pos *= sample distance
   // Inc(pValue, pChanGrp.CNNumber - 1);             // todo, +Recordsize -> pValue = Byte-Pointer
end;


procedure TFrameViewData.ChkReadOnSlideClick(Sender: TObject);
begin
   ReadOnSlide := ChkReadOnSlide.Checked;
end;


procedure TFrameViewData.BtnSelectFileClick(Sender: TObject);
begin
   // MdfFileDlg.InitialDir := DmSettings.FolderNameMDF4;
   // MdfFileDlg.Filter := 'ASAM MDF4 files (*.mf4)|*.mf4|';
   // MdfFileDlg.Execute;
   // if MdfFileDlg.Filename = '' then
   // begin
   // Filename := '';
   // Exit;
   // end;
   // Filename := MdfFileDlg.Filename;
   InitFile;
end;


procedure TFrameViewData.BtnSelDefFileClick(Sender: TObject);
begin
   Filename := FilenameDefault;
   InitFile;
end;


procedure TFrameViewData.BtnUnDoZoomClick(Sender: TObject);
begin
   if ViewChart.SeriesCount > 0 then
      ViewChart.UndoZoom;
end;


procedure TFrameViewData.ChkCursorClick(Sender: TObject);
begin
   if ChkCursor.Checked then
   begin
      CursorOldX := -1;
      CursorOldY := -1;
      RectTool.ParentChart := ViewChart;
      RectTool.Position := ppRightTop;
      RectTool.Active := True;
      CursorTool.Active := True;
      CursorTool.ParentChart := ViewChart;
      CursorTool.Series := ViewChart.Series[0];
      CursorTool.XValue := 0.5 * (ViewChart.BottomAxis.Maximum - ViewChart.BottomAxis.Minimum) + ViewChart.BottomAxis.Minimum;
      CursorView_UpdateValues(ViewChart.BottomAxis.CalcPosValue(CursorTool.XValue));
   end else begin
      CursorTool.Series := nil;
      CursorTool.Active := False;
      RectTool.ParentChart := nil;
      RectTool.Active := False;
   end;
end;


procedure TFrameViewData.ChkMeasDateClick(Sender: TObject);
var
   pSerie: ^TChartSeries;
begin
   if NumSignals < 1 then
      Exit;

   if ChkMeasDate.Checked then
   begin
      BottomAxisTimeType := xtDate;
      ViewChart.BottomAxis.DateTimeFormat := 'hh:mm:ss,zzz' + #13 + 'ddddd';
      for var SignalIdx := 0 to NumSignals - 1 do
      begin
         pSerie := @Signals[SignalIdx].Serie;
         pSerie.XValues.DateTime := True;
      end;
   end else begin
      BottomAxisTimeType := xtMeas;
      ViewChart.BottomAxis.AxisValuesFormat := '#,##0.#';
      for var SignalIdx := 0 to NumSignals - 1 do
      begin
         pSerie := @Signals[SignalIdx].Serie;
         pSerie.XValues.DateTime := False;
      end;
   end;
   Self.BtnReadDataClick(Sender);

{$IFDEF USE_SCROLLPAGER}
   // repeat scrollpager initialisation from 'InitFile' procedure
   // changes to the series time base (date+time or measurement time) are not automatically applied
   // to the scrollpager's sub chart and axis/axes

   ScrollPagerTool.Visible := False;
   ScrollPagerTool.SubChartTChart.RemoveAllSeries;

   // select series from main chart to be shown as preview in ScrollPager's sub chart
   ScrollPagerTool.AddSeries(ViewChart[0]);
   ScrollPagerTool.Series.Visible := True;

   // set left axis of series of main chart to be used as left axis of series in Scrollpager's sub chart
   ScrollPagerTool.SubChartTChart.Axes.Left.Assign(Signals[0].Serie.GetVertAxis);
   ScrollPagerTool.SubChartTChart.LeftAxis.Visible := True;

   // if BottomAxisTimeType = xtDate <this> else <that>
   ScrollPagerTool.SubChartTChart.BottomAxis.DateTimeFormat := ViewChart.BottomAxis.DateTimeFormat;
   ScrollPagerTool.SubChartTChart.BottomAxis.AxisValuesFormat := ViewChart.BottomAxis.AxisValuesFormat;

   ScrollPagerTool.SubChartTChart.BottomAxis.Automatic := False;
   ScrollPagerTool.SubChartTChart.BottomAxis.AutomaticMinimum := True;
   ScrollPagerTool.SubChartTChart.BottomAxis.AutomaticMaximum := True;

   // show labels with numeric value of start/end position's Y value of series
   ScrollPagerTool.ColorBandTool.StartLine.AnnotationValue := True;
   ScrollPagerTool.ColorBandTool.EndLine.AnnotationValue := True;

   // determine beginning/ending of Scrollpager's handle
   var
      XStart, XEnd: Double;
   XStart := ViewChart[0].XValue[0];                        // first x value of first series in main chart
   XEnd := ViewChart[0].XValue[Signals[0].Serie.Count - 1]; // last value/record

   ScrollPagerTool.ColorBandTool.StartValue := XStart;
   ScrollPagerTool.ColorBandTool.EndValue := XEnd;

   ScrollPagerTool.SubChartTChart.BottomAxis.Increment := (XEnd - XStart) / 5.0;


   ScrollPagerTool.Visible := True;
{$ENDIF}
   // ---
   // Wenn man hier nur die X-Werte ndert, dann dauert es lnger bis das Update durch ist (>1sec)
   // als wenn man alles neu aus der Datei liest und die Serien komplett neu aufbaut?!
   // >>> Warum!?
   //
   // pSerie.BeginUpdate;
   // // if NumberRecords <> pSerie.XValues.Count then
   // // ImpossibleCase_Hopefully;
   // for var SampleIdx := 0 to NumberRecords - 1 do
   // pSerie.XValue[SampleIdx] := IncMilliSecond(MdfFileStartTime, round((StartRecord + SampleIdx) * SampleDistance * 1000));
   // pSerie.EndUpdate;
   // pSerie.ParentChart := ViewChart;
   // pSerie.BeginUpdate;
   // for var SampleIdx := 0 to NumberRecords - 1 do
   // pSerie.XValue[SampleIdx] := (StartRecord + SampleIdx) * SampleDistance;
   // pSerie.EndUpdate;
   // pSerie.ParentChart := ViewChart;
end;


procedure TFrameViewData.BtnStartClick(Sender: TObject);
begin
   if StartRecord <> 0 then
   begin
      StartRecord := 0;
      EdStartRecord.Text := StartRecord.ToString;
      RecordSlider.SliderPosition := StartRecord;
      Self.BtnReadDataClick(Sender);
   end;
end;


procedure TFrameViewData.BtnEndClick(Sender: TObject);
begin
   if StartRecord <> MdfNumberRecords - NumberRecords then
   begin
      StartRecord := MdfNumberRecords - NumberRecords;
      EdStartRecord.Text := StartRecord.ToString;
      RecordSlider.SliderPosition := StartRecord;
      Self.BtnReadDataClick(Sender);
   end;
end;


procedure TFrameViewData.BtnNextClick(Sender: TObject);
var
   Start: UInt32;
begin
   Start := StartRecord + NumberRecords;
   if Start > MdfNumberRecords - NumberRecords then
      Start := MdfNumberRecords - NumberRecords;
   if Start <> StartRecord then
   begin
      StartRecord := Start;
      EdStartRecord.Text := StartRecord.ToString;
      RecordSlider.SliderPosition := StartRecord;
      Self.BtnReadDataClick(Sender);
   end;
end;


procedure TFrameViewData.BtnPrevClick(Sender: TObject);
var
   Start: UInt32;
begin
   if StartRecord < NumberRecords - 1 then
      Start := 0
   else
      Start := StartRecord - NumberRecords;
   if Start <> StartRecord then
   begin
      StartRecord := Start;
      EdStartRecord.Text := StartRecord.ToString;
      RecordSlider.SliderPosition := StartRecord;
      Self.BtnReadDataClick(Sender);
   end;
end;


// Slider - while moving
procedure TFrameViewData.OnChangePosition(Sender: TObject);
var
   MaxPos: Double;
begin
   if not(Sender = RecordSlider) then
      Exit;
   // if pMdfMeas = nil then
   // Exit;
   MaxPos := MdfNumberRecords - NumberRecords;
   if RecordSlider.SliderPosition > MaxPos then
   begin
      RecordSlider.SliderPosition := MaxPos;
      RecordSlider.SliderSpan := NumberRecords;
   end;
   StartRecord := Trunc(RecordSlider.SliderPosition);
   EdStartRecord.Text := StartRecord.ToString;

   if ReadOnSlide then
   begin
      Self.BtnReadDataClick(Self);
      RecordSlider.Repaint; // force slider graphics to follow mouse-pointer
      ViewChart.Repaint;
   end;
end;


// Slider - after position settled
procedure TFrameViewData.OnSetPosition(Sender: TObject);
begin
   if not(Sender = RecordSlider) then
      Exit;
   // if pMdfMeas = nil then
   // Exit;
   StartRecord := Trunc(RecordSlider.SliderPosition);
   Self.BtnReadDataClick(Sender);
end;


// EditFields
procedure TFrameViewData.OnEditExit(Sender: TObject);
var
   Start, Number: UInt32;
   Tick: Double;
   Update: Boolean;
   Scale: TChartAxis;
begin
   Update := False;
   if (Sender = EdStartRecord) then
   begin
      Start := StrToInt(EdStartRecord.Text);
      if Start > MdfNumberRecords - NumberRecords then
         Start := MdfNumberRecords - NumberRecords;
      StartRecord := Start;
      EdStartRecord.Text := StartRecord.ToString;
      RecordSlider.SliderPosition := StartRecord;
      Update := True;
   end;
   if (Sender = EdNumberRecords) then
   begin
      Number := StrToInt(EdNumberRecords.Text);
      if Number < 100 then
         Number := min(100, MdfNumberRecords - StartRecord);
      if Number + StartRecord > MdfNumberRecords then
         Number := MdfNumberRecords - StartRecord;
      NumberRecords := Number;
      EdNumberRecords.Text := NumberRecords.ToString;
      RecordSlider.SliderSpan := NumberRecords;
      SliderSpan := NumberRecords;
      Update := True;
   end;
   if (Sender = EdAxisMax) then
   begin
      if IdxSelectedScale > -1 then
      begin
         Scale := ViewChart.CustomAxes[IdxSelectedScale];
         Tick := StrToFloat(EdAxisMax.Text);
         if Tick > Scale.Minimum then
            Scale.Maximum := Tick;
         EdAxisMax.Text := FloatToStrF(Tick, ffFixed, 10, 3);

         // Tick := StrToFloat(EdAxisMax.Text);
         // if Tick > Scales[IdxSelectedScale].Minimum then
         // Scales[IdxSelectedScale].Maximum := Tick;
         // EdAxisMax.Text := FloatToStrF(Tick, ffFixed, 10, 3);
      end;
   end;
   if (Sender = EdAxisMin) then
   begin
      if IdxSelectedScale > -1 then
      begin
         Scale := ViewChart.CustomAxes[IdxSelectedScale];
         Tick := StrToFloat(EdAxisMin.Text);
         if Tick < Scale.Maximum then
            Scale.Minimum := Tick;
         EdAxisMin.Text := FloatToStrF(Tick, ffFixed, 10, 3);

         // Tick := StrToFloat(EdAxisMin.Text);
         // if Tick < Scales[IdxSelectedScale].Maximum then
         // Scales[IdxSelectedScale].Minimum := Tick;
         // EdAxisMin.Text := FloatToStrF(Tick, ffFixed, 10, 3);
      end;
   end;

   if Update then
   begin
      if (Sender = EdStartRecord) or (Sender = EdNumberRecords) then
      begin
         SetBufferLength(MdfRecordSize * NumberRecords); // chunkSize in DLL
         Self.BtnReadDataClick(Sender);                  // erstmal nur so
      end;
   end;
end;


procedure TFrameViewData.FormKeyPress(Sender: TObject; var Key: Char);
begin
   if CharInSet(Key, [#13]) then
   begin
      OnEditExit(Sender);
      Exit;
   end;
   if (Sender = EdStartRecord) or (Sender = EdNumberRecords) then
   begin
      // no negative numbers
      if not CharInSet(Key, ['0' .. '9', #8]) then // #8=Backspace
         Key := #0;                                // do not pass this key
   end;
end;


// Chart
procedure TFrameViewData.OnChartAllowScroll(Sender: TChartAxis; var AMin, AMax: TAxisValue; var AllowScroll: Boolean);
begin
   // Example:
   // data of 100 records loaded, zoom-window displays data of 10 records
   // --> TChart's build-in scoll (hold RMB and drag) works automatically within the loaded data
   // --> cannot scroll beyond the edges (needs loading/streaming of new data from file)
   if Sender = ViewChart.BottomAxis then
   begin
      if AMin < 0 then
         AllowScroll := False;
      if AMax > MdfNumberRecords then
         AllowScroll := False;

      // if allow_to_scroll_only_in_records_loaded then
      if AMin < ScrollMinX then
         AllowScroll := False;
      if AMax > ScrollMaxX then
         AllowScroll := False;
   end;
end;


procedure TFrameViewData.OnChartClickSeries(Sender: TCustomChart; ASerie: TChartSeries; ValueIndex: Integer; Button: TMouseButton; Shift: TShiftState;
   X, Y: Integer);
var
   GridRelocationDone: UInt32;
   Scale: TChartAxis;
begin
   if CursorTool.Active then
   begin
      CursorTool.Series := ASerie;
      CursorTool.XValue := CursorTool.Series.XScreenToValue(X);
      CursorTool.YValue := CursorTool.Series.YScreenToValue(Y); // todo, fix the little offset if serie was not clicked perfectly
      CursorView_UpdateValues(X);
   end;

   GridRelocationDone := 0;
   // assumption: series and scale idx match
   for var Idx := 0 to NumSignals - 1 do
   begin
      Scale := ViewChart.CustomAxes[Idx];

      // hide current scale's grid
      if Scale.Grid.Visible = True then
      begin
         Scale.Grid.Visible := False;
         Inc(GridRelocationDone);
      end;
      // show grid for clicked series
      if Signals[Idx].Serie = ASerie then
      begin
         Scale.Grid.Visible := True;
         Inc(GridRelocationDone);
      end;
      if GridRelocationDone = 2 then
         break;
   end;

   // for var Idx := 0 to NumSignals - 1 do
   // begin
   // // hide current scale's grid
   // if Scales[Idx].Grid.Visible = True then
   // begin
   // Scales[Idx].Grid.Visible := False;
   // Inc(GridRelocationDone);
   // end;
   // // show grid for clicked series
   // if Signals[Idx].Serie = ASerie then
   // begin
   // Scales[Idx].Grid.Visible := True;
   // Inc(GridRelocationDone);
   // end;
   // if GridRelocationDone = 2 then
   // break;
   // end;


   // test
   var
      i: Integer;
   i := ViewChart.CustomAxes[0].Texts.MarginToAxis;
   ViewChart.CustomAxes[0].Texts.MarginToAxis := ViewChart.CustomAxes[0].Texts.MarginToAxis + 20;
   i := ViewChart.CustomAxes[0].Texts.MarginToAxis;
   Inc(i); // breakpoint
   ViewChart.Update;
end;


procedure TFrameViewData.OnChartClickAxis(Sender: TCustomChart; Axis: TChartAxis; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
   KeyState: TShiftState;
   AxMin, AxMax: Double;
   Scale: TChartAxis;
begin
   KeyState := Shift * [ssShift];
   for var Idx := 0 to NumScales - 1 do
   begin
      Scale := ViewChart.CustomAxes[Idx];

      if Scale = Axis then
      begin
         if Idx = Cardinal(IdxSelectedScale) then
         begin
            IdxSelectedScale := -1;
            Exit;
         end;
         IdxSelectedScale := Idx;
         if ssShift in KeyState then
         begin
            AxMin := StrToFloat(EdAxisMin.Text);
            AxMax := StrToFloat(EdAxisMax.Text);
            Scale.SetMinMax(AxMin, AxMax);
         end else begin
            AxMin := Scale.Minimum;
            AxMax := Scale.Maximum;
            EdAxisMin.Text := FloatToStrF(AxMin, ffFixed, 10, 3);
            EdAxisMax.Text := FloatToStrF(AxMax, ffFixed, 10, 3);
         end;
         break;
      end;
   end;

   // KeyState := Shift * [ssShift];
   // for var Idx := 0 to NumScales - 1 do
   // begin
   // if Scales[Idx] = Axis then
   // begin
   // if Idx = Cardinal(IdxSelectedScale) then
   // begin
   // IdxSelectedScale := -1;
   // Exit;
   // end;
   // IdxSelectedScale := Idx;
   // if ssShift in KeyState then
   // begin
   // AxMin := StrToFloat(EdAxisMin.Text);
   // AxMax := StrToFloat(EdAxisMax.Text);
   // Scales[Idx].SetMinMax(AxMin, AxMax);
   // end else begin
   // AxMin := Scales[Idx].Minimum;
   // AxMax := Scales[Idx].Maximum;
   // EdAxisMin.Text := FloatToStrF(AxMin, ffFixed, 10, 3);
   // EdAxisMax.Text := FloatToStrF(AxMax, ffFixed, 10, 3);
   // end;
   // break;
   // end;
   // end;
end;


procedure TFrameViewData.OnClickLegend(Sender: TCustomChart; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
   Serie: TChartSeries;
   idxClicked, Offset: Integer;
   Delta, min, max: Double;
   KeyState: TShiftState;
   Scale: TChartAxis;
begin
   if Button <> TMouseButton.mbLeft then
      Exit;

   // Shift & Left-click: left axis min/max ticks = clicked serie min/max
   KeyState := Shift * [ssShift];
   if ssShift in KeyState then
   begin
      idxClicked := ViewChart.Legend.Clicked(X, Y);
      if idxClicked >= 0 then
      begin
         Serie := ViewChart.Series[idxClicked];
         Delta := Serie.MaxYValue - Serie.MinYValue;
         min := System.round(Serie.MinYValue - (Delta * 0.1));
         max := System.round(Serie.MaxYValue + (Delta * 0.1));
         ViewChart.LeftAxis.SetMinMax(min, max);
         EdAxisMin.Text := FloatToStrF(min, ffFixed, 10, 3);
         EdAxisMax.Text := FloatToStrF(max, ffFixed, 10, 3);
         Serie.Visible := True; // undo auto-hiding when clicking the legend
      end;
   end;

   // position of scales set in percent of chart width (left=0%)
   // PositionUnits = muPercent
   // Offset := 0;
   // for var Idx := 0 to NumSignals - 1 do
   // begin
   // if Signals[Idx].Serie.Visible then
   // begin
   // Scales[Idx].PositionPercent := ScalePosFactor * Offset + ScalePosOffset;
   // Inc(Offset);
   // end;
   // end;

   Offset := 0;
   for var Idx := 0 to NumSignals - 1 do
   begin
      if Signals[Idx].Serie.Visible then
      begin
         Scale := ViewChart.CustomAxes[Idx];

         Scale.PositionPercent := ScalePosFactor * Offset + ScalePosOffset;
         Inc(Offset);
      end;
   end;

   ViewChart.BottomAxis.StartPosition := ScalePosFactor * (Offset - 1) + ScalePosOffset;

   // --- does not work as intended (don't know why)
   //
   // Ctrl & Left-click: unselect all series but the clicked one
   // KeyState := Shift * [ssCtrl];
   // if ssCtrl in KeyState then
   // begin
   // idxClicked := ViewChart.Legend.Clicked(X, Y);
   // begin
   // for var Idx := 0 to ViewChart.SeriesCount - 1 do
   // begin
   // Serie := ViewChart.Series[idxClicked];
   // if Idx = idxClicked then
   // begin
   // // Delta := Serie.MaxYValue - Serie.MinYValue;
   // // min := System.round(Serie.MinYValue - (Delta * 0.1));
   // // max := System.round(Serie.MaxYValue + (Delta * 0.1));
   // // ViewChart.LeftAxis.SetMinMax(min, max);
   // // EdAxisMin.Text := FloatToStrF(min, ffFixed, 10, 3);
   // // EdAxisMax.Text := FloatToStrF(max, ffFixed, 10, 3);
   // Serie.Visible := True;
   // // undo auto-hiding when clicking the legend
   // end else begin
   // Serie.Visible := False;
   // end;
   // end;
   // end;
   // end;
end;


procedure TFrameViewData.OnCursorToolChange(Sender: TCursorTool; X, Y: Integer; const XValue, YValue: Double; Series: TChartSeries; ValueIndex: Integer);
var
   CursorMovedX, CursorMovedY: Boolean;
begin
   if (ChkCursor.Checked = False) or (ValueIndex < 0) then
      Exit;

   CursorMovedX := False;
   if X <> CursorOldX then
   begin
      CursorOldX := X;
      CursorMovedX := True;
   end;

   CursorMovedY := False;
   if Y <> CursorOldX then
   begin
      CursorOldY := Y;
      CursorMovedY := True;
   end;

   if CursorMovedY and not CursorMovedX then
      Exit;

   CursorView_UpdateValues(X);
end;


procedure TFrameViewData.OnZoom(Sender: TObject);
var
   Delta: Double;
   Decimals: Integer;
   ValueFormat: string;
   Scale: TChartAxis;
begin
   // y
   for var Idx := 0 to NumScales - 1 do
   begin
      Scale := ViewChart.CustomAxes[Idx];

      // apply zoom level to custom axes
      // custom axes do *not* zoom automatically as the default ones do
      Scale.SetMinMax(Scale.CalcPosPoint(ViewChart.Zoom.Y1), Scale.CalcPosPoint(ViewChart.Zoom.Y0));

      Delta := Scale.Maximum - Scale.Minimum;
      Scale.Increment := Delta / 4.0;
      ValueFormat := '#,##0.';
      Decimals := 1;
      while ((Delta / Scale.Items.Count) * (Decimals * 10) < 1) and (Decimals < 13) do
      begin
         Inc(Decimals);
         ValueFormat := ValueFormat + '#';
      end;
      // 0 at pos x: show at least this amount of decimals, trailing zeros
      // '1.1' --> xyz.##0 --> '1.100'
      ValueFormat := ValueFormat + '0';
      Scale.AxisValuesFormat := ValueFormat;
   end;

   // x
   Delta := ViewChart.BottomAxis.Maximum - ViewChart.BottomAxis.Minimum;
   ViewChart.BottomAxis.Increment := Delta / 5.0;
   case BottomAxisTimeType of
      xtMeas:
         begin
            ValueFormat := '#,##0.#';
            Decimals := 1;
            while ((Delta / ViewChart.BottomAxis.Items.Count) * (Decimals * 10) < 1) and (Decimals < 13) do
            begin
               Inc(Decimals);
               ValueFormat := ValueFormat + '#';
            end;
            ViewChart.BottomAxis.AxisValuesFormat := ValueFormat;
         end;
      xtDate:
         begin
            ViewChart.BottomAxis.DateTimeFormat := 'hh:mm:ss.zzz' + #13 + 'ddddd';
         end;
   end;

   // // y
   // for var Idx := 0 to NumScales - 1 do
   // begin
   // // apply zoom level to custom axes
   // // custom axes do *not* zoom automatically as the default ones do
   // Scales[Idx].SetMinMax(Scales[Idx].CalcPosPoint(ViewChart.Zoom.Y1), Scales[Idx].CalcPosPoint(ViewChart.Zoom.Y0));
   //
   // Delta := Scales[Idx].Maximum - Scales[Idx].Minimum;
   // Scales[Idx].Increment := Delta / 4.0;
   // ValueFormat := '#,##0.';
   // Decimals := 1;
   // while ((Delta / Scales[Idx].Items.Count) * (Decimals * 10) < 1) and (Decimals < 13) do
   // begin
   // Inc(Decimals);
   // ValueFormat := ValueFormat + '#';
   // end;
   // // 0 at pos x: show at least this amount of decimals, trailing zeros
   // // '1.1' --> xyz.##0 --> '1.100'
   // ValueFormat := ValueFormat + '0';
   // Scales[Idx].AxisValuesFormat := ValueFormat;
   // end;
   //
   // // x
   // Delta := ViewChart.BottomAxis.Maximum - ViewChart.BottomAxis.Minimum;
   // ViewChart.BottomAxis.Increment := Delta / 5.0;
   // case BottomAxisTimeType of
   // xtMeas:
   // begin
   // ValueFormat := '#,##0.#';
   // Decimals := 1;
   // while ((Delta / ViewChart.BottomAxis.Items.Count) * (Decimals * 10) < 1) and (Decimals < 13) do
   // begin
   // Inc(Decimals);
   // ValueFormat := ValueFormat + '#';
   // end;
   // ViewChart.BottomAxis.AxisValuesFormat := ValueFormat;
   // end;
   // xtDate:
   // begin
   // ViewChart.BottomAxis.DateTimeFormat := 'hh:mm:ss.zzz' + #13 + 'ddddd';
   // end;
   // end;
end;


procedure TFrameViewData.OnUndoZoom(Sender: TObject);
var
   Delta: Double;
   Decimals: Integer;
   ValueFormat: string;
   Scale: TChartAxis;
begin
   // y
   for var Idx := 0 to NumScales - 1 do
   begin
      Scale := ViewChart.CustomAxes[Idx];

      Scale.SetMinMax(-10.0, 10.0);

      Delta := Scale.Maximum - Scale.Minimum;
      Scale.Increment := Delta / 4.0;

      ValueFormat := '#,##0.';
      Decimals := 1;
      while ((Delta / Scale.Items.Count) * (Decimals * 10) < 1) and (Decimals < 13) do
      begin
         Inc(Decimals);
         ValueFormat := ValueFormat + '#';
      end;
      ValueFormat := ValueFormat + '0';
      Scale.AxisValuesFormat := ValueFormat;
   end;

   // // y
   // for var Idx := 0 to NumScales - 1 do
   // begin
   // Scales[Idx].SetMinMax(-10.0, 10.0);
   //
   // Delta := Scales[Idx].Maximum - Scales[Idx].Minimum;
   // Scales[Idx].Increment := Delta / 4.0;
   //
   // ValueFormat := '#,##0.';
   // Decimals := 1;
   // while ((Delta / Scales[Idx].Items.Count) * (Decimals * 10) < 1) and (Decimals < 13) do
   // begin
   // Inc(Decimals);
   // ValueFormat := ValueFormat + '#';
   // end;
   // ValueFormat := ValueFormat + '0';
   // Scales[Idx].AxisValuesFormat := ValueFormat;
   // end;
   // x
   Delta := ViewChart.BottomAxis.Maximum - ViewChart.BottomAxis.Minimum;
   ViewChart.BottomAxis.Increment := Delta / 5.0;
   case BottomAxisTimeType of
      xtMeas:
         begin
            ValueFormat := '#,##0.#';
            Decimals := 1;
            while ((Delta / ViewChart.BottomAxis.Items.Count) * (Decimals * 10) < 1) and (Decimals < 13) do
            begin
               Inc(Decimals);
               ValueFormat := ValueFormat + '#';
            end;
            ViewChart.BottomAxis.AxisValuesFormat := ValueFormat;
         end;
      xtDate:
         begin
            ViewChart.BottomAxis.DateTimeFormat := 'hh:mm:ss.zzz' + #13 + 'ddddd';
         end;
   end;
end;


procedure TFrameViewData.OnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
   // ...
end;


procedure TFrameViewData.OnMouseUp(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
   // ...
end;



// Problem:
// (1) Auf Y-Achse scrollende Beschriftungen werden oben/unten ber Rand hinaus gezeichnet.
// (2) Gitterlinien von linker Skala werden in anderen Skalen gezeichnet.
//
// Reihenfolge OnDraw_XYZ: Chart --> Axes --> Series
//
// Funktioniert so nicht:
// Clip-Rechteck ist Canvas, beschrnkt auf Top/Bottom der vertikalen Achsen:
// X-Achse wird im weggclipten Bereich (nicht) gezeichnet
// >> Es gibt kein OnBefore_VertikaleAchseZeichnen / OnBefore_HorizontaleAchseZeichnen.
// >> Es gibt kein OnBefore/After_GitterZeichnen, um einen Clipping-Bereich zu definieren
// >> Weiss nicht, wie man das lsen soll.


procedure TFrameViewData.OnBeforeDrawAxes(Sender: TObject);
// var
// ClipRect: TRect;
// Axis: TChartAxis;
begin
   if NumSignals < 1 then
      Exit;

   // Axis := TChartAxis(Sender);
   // if Signals[0].Serie.CustomVertAxis = Axis then
   // begin
   // if not Axis.Visible then // test
   // Exit;                 // "
   // end;
end;


procedure TFrameViewData.OnBeforeDrawAxis(Sender: TObject);
var
   ClipRect: TRect;
   // Axis: TChartAxis;
begin
   if NumSignals < 1 then
      Exit;
end;



procedure TFrameViewData.OnBeforeDrawSeries(Sender: TObject);
var
   ClipRect: TRect;
   Serie: TChartSeries;
begin
   if NumSignals < 1 then
      Exit;

   Serie := TChartSeries(Sender);
   if not Serie.Visible then
      Exit;

   // clip to area spanned by signal's axes
   ClipRect.Top := Serie.GetVertAxis.IStartPos;
   ClipRect.Left := Serie.GetHorizAxis.IStartPos; // about to be modified below
   ClipRect.Right := Serie.GetHorizAxis.IEndPos;
   ClipRect.Bottom := Serie.GetVertAxis.IEndPos;

   // determine the rightmost scale of visible series
   // if ViewChart.CustomAxes.Count > 0 then // fixme, a serie is visible => Count>0
   // begin
   ClipRect.Left := Serie.GetHorizAxis.IStartPos;
   for var Idx := NumSignals - 1 downto 0 do // to
   begin
      if not Signals[Idx].Serie.Visible then
         continue;
      ClipRect.Left := Signals[Idx].Serie.GetHorizAxis.IStartPos;
      break;
   end;
   // end;
   ViewChart.Canvas.Clip(ClipRect);
end;


procedure TFrameViewData.OnAfterDrawSeries(Sender: TObject);
begin
   ViewChart.Canvas.UnClip;
end;


procedure TFrameViewData.OnBeforeDrawChart(Sender: TObject);
begin
   if NumSignals < 1 then
      Exit;
   ViewChart.Canvas.UnClip;
end;


// does not work: access violation on 'Draw' in 'PaintWindow' (VCL.Controls), why? do not know...
// vertical cursor line: draw ellipse at intersection point to a serie
procedure TFrameViewData.OnChartAfterDraw(Sender: TObject);
// var
// X, Y: Integer;
// pSerie: ^TFastLineSeries;
begin
   if NumSignals < 1 then
      Exit;
   ViewChart.Canvas.UnClip;

   // ScrollPagerTool.SetUpScrollPager;

   // if CursorTool.Active then
   // begin
   // X := ViewChart.BottomAxis.CalcXPosValue(CursorXValue);
   // for var Idx := 0 to NumSignals - 1 do
   // begin
   // Y := Scales[Idx].CalcPosValue(Signals[Idx].Serie.YValue[X]);
   // ViewChart.Canvas.Brush.Color := Signals[Idx].Serie.Color;
   // // TLineSeries(Signals[Idx].Serie).Pointer.Draw(X, Y);      //  <-- !
   // end;
   // end;
end;


procedure TFrameViewData.OnAxisDrawLabel(Sender: TChartAxis; var X, Y, Z: Integer; var Text: string; var DrawLabel: Boolean);
// var
// S: string;
begin
   // funktioniert nicht, da eine serie nicht zwei achsen haben kann (hidden grid scale)
   // if Sender = Scales[NumScales] then
   // begin
   // DrawLabel := False;
   // Exit;
   // end;

   // X := X + 500; // no change of position, why...?!
   // Y := Y + 500;
   // S := Text + #13;
   // S := S + X.ToString + #13;
   // S := S + Y.ToString + #13;
   // S := S + Z.ToString + #13;
   // ShowMessage(S);
   // ...
   DrawLabel := True;
end;


procedure TFrameViewData.OnBottomAxisDrawLabel(Sender: TChartAxis; var X, Y, Z: Integer; var Text: string; var DrawLabel: Boolean);
// var
// S: string;
begin
   // S := Text + #13;
   // S := S + X.ToString + #13;
   // S := S + Y.ToString + #13;
   // S := S + Z.ToString + #13;
   // ShowMessage(S);
   // ...
   DrawLabel := True;
end;


procedure TFrameViewData.OnGetAxisLabel(Sender: TChartAxis; Series: TChartSeries; ValueIndex: Integer; var LabelText: string);
var
   S: string;
begin
   // LabelText := 'hallo';

   // if (Sender = ViewChart.BottomAxis) then
   // begin
   // S := LabelText + #13;
   // S := S + ValueIndex.ToString; //always equal to -1, why...?!
   //
   // ShowMessage(S);
   // end;
   // ...
end;


// wozu ist diese funktion gut?!
procedure TFrameViewData.OnGetNextAxisLabel(Sender: TChartAxis; LabelIndex: Integer; var LabelValue: Double; var Stop: Boolean);
// var
// S: string;
begin
   // if (Sender = ViewChart.BottomAxis) then
   // begin
   // S := LabelIndex.ToString + #13;
   // S := S + LabelValue.ToString;
   // // LabelValue := 0;
   //
   // if LabelIndex > 1000 then
   // ShowMessage(S);
   //
   // Stop := False;
   // end
   // else
   // Stop := True;
   // ...
end;


initialization


FilenameDefault := 'C:\Users\Public\Documents\Demeas\DataStorage\DefaultData.mf4';

StartRecord := 0;
NumberRecords := 100000;
ByteBuffer := nil;
BufferLength := 0;
MdfRecordSize := 0;
MdfNumberRecords := 0;
GridColor := TColor(RGB(50, 50, 50));
BackgroundColor := TColor(RGB(47, 47, 47));
ScaleColor := TColor(RGB(150, 150, 150));
ScaleTextColor := TColor(RGB(150, 150, 150));;


end.

// -- Halde

// ---

// pDataGrp := @(pMdfMeas.DataGroups[0]);                  // param
// pChanGrp := @(pMdfMeas.DataGroups[0].ChannelGroups[0]); // demeas creates sorted Mdf files: single CG only
//
// NumSignals := min(MAX_SIGNALS, pChanGrp.CNNumber - 1);  // skip virtual time channel
// SetLength(Signals, NumSignals);
// for var Idx := 0 to NumSignals - 1 do
// begin
// Signals[Idx].MdfChanIdx := Idx + 1;                           // 0 = virtual time channel
// Signals[Idx].Offset := pChanGrp.Channels[Idx + 1].ByteOffset;
// Signals[Idx].Bytes := pChanGrp.Channels[Idx + 1].ByteSize;
// Signals[Idx].DataType := pChanGrp.Channels[Idx + 1].DataType; // Mdf data type, e.g. CN_D_FLOAT_LE
// Signals[Idx].Serie := TFastLineSeries.Create(ViewChart);
// Signals[Idx].Serie.Title := PWChar(pChanGrp.Channels[Idx + 1].ChanName);
// Signals[Idx].Serie.Pen.Width := 1;
// Signals[Idx].Serie.BeforeDrawValues := OnBeforeDrawSeries;
// Signals[Idx].Serie.AfterDrawValues := OnAfterDrawSeries;
//
// TFastLineSeries(Signals[Idx].Serie).FastPen := True;
// TFastLineSeries(Signals[Idx].Serie).DrawAllPoints := False; // do not draw multiple values to the same x-position
// TFastLineSeries(Signals[Idx].Serie).ClickTolerance := 2;
//
// case BottomAxisTimeType of
// xtMeas:
// Signals[Idx].Serie.XValues.DateTime := False;
// xtDate:
// Signals[Idx].Serie.XValues.DateTime := True;
// end;
// end;
//
// if ViewChart.CustomAxes.Count > 0 then
// begin
// while ViewChart.CustomAxes.Count > 0 do
// ViewChart.CustomAxes.Delete(0);
// NumScales := 0;
// end;
//
// NumScales := NumSignals;
// SetLength(Scales, NumScales);
// NumScrollTools := NumSignals;
// SetLength(ScrollTools, NumScrollTools);
// for var Idx := 0 to NumScales - 1 do
// begin
// ViewChart.CustomAxes.Add();
// Scales[Idx] := ViewChart.CustomAxes[Idx];
// Scales[Idx].Grid.Hide;
// Scales[Idx].Grid.Color := GridColor;
// Scales[Idx].MinorTickCount := 1;
// Scales[Idx].SetMinMax(-10.0, 10.0);
// Scales[Idx].Increment := 2.5;
// Scales[Idx].Title.Text := '';
// Scales[Idx].Title.Position := tpCenter;
// Scales[Idx].Title.Alignment := taRightJustify;
// Scales[Idx].Title.Font.Name := DmSettings.UiFontName;
// Scales[Idx].Title.Font.SizeFloat := -MulDiv(DmSettings.ScaleFontSize, CurrentPPI, 72);
// Scales[Idx].PositionPercent := Idx * ScalePosFactor + ScalePosOffset; // vor muldiv fontsize: 3% der chartbreite fr eine achse
// Scales[Idx].MinimumOffset := 10;
// Scales[Idx].MaximumOffset := 10;
// Scales[Idx].LabelsFont.Name := DmSettings.UiFontName;
// Scales[Idx].LabelsFont.SizeFloat := -MulDiv(DmSettings.ScaleFontSize, CurrentPPI, 72);
// Scales[Idx].LabelsFont.Quality := fqClearType;
// Scales[Idx].LabelsAngle := 90;
// Scales[Idx].LabelsFont.Color := ScaleTextColor;
// Scales[Idx].Axis.Color := ScaleColor;
// Scales[Idx].Ticks.Color := ScaleColor;
// // Das ndert den Abstand des Kanalnames(!) zur Achse,
// // aber nicht die Gre (Size!) oder den Abstand der Tickbeschriftung(!) zur Achse...!?
// // Scales[idx].LabelsSize := 2;
//
// if Idx = 0 then
// begin
// Scales[Idx].Grid.Show;
// Scales[Idx].Grid.Width := 2;
// end;
//
// Delta := Scales[Idx].Maximum - Scales[Idx].Minimum;
// ValueFormat := '#,##0.';
// Decimals := 1;
// while ((Delta / Scales[Idx].Items.Count) * (Decimals * 10) < 1) and (Decimals < 13) do
// begin
// Inc(Decimals);
// ValueFormat := ValueFormat + '#';
// end;
// // 0 at pos x: show at least this amount of decimals, trailing zeros
// // '1.1' --> xyz.##0 --> '1.100'
// ValueFormat := ValueFormat + '0';
// Scales[Idx].AxisValuesFormat := ValueFormat;
//
// ScrollTools[Idx] := TAxisScrollTool.Create(ViewChart);
// ScrollTools[Idx].Axis := Scales[Idx];
//
// // attach scale to serie
// Signals[Idx].Serie.CustomVertAxis := Scales[Idx];
// Scales[Idx].Title.Text := Signals[Idx].Serie.Title;
// end;
//
// ViewChart.BottomAxis.StartPosition := ScalePosFactor * (NumScales - 1) + ScalePosOffset;
// ViewChart.BottomAxis.EndPosition := 100.0;
//
// MdfSampleDistance := pChanGrp.SampleDistance;
// MdfNumberRecords := pChanGrp.RecordNumber;
// MdfRecordSize := pChanGrp.RecordSize;
//
// StartRecord := 0;
// NumberRecords := 100000;
// if pChanGrp.RecordNumber < 100000 then
// NumberRecords := MdfNumberRecords;
//
// SetBufferLength(MdfRecordSize * NumberRecords);
//
// EdStartRecord.Text := StartRecord.ToString;
// EdNumberRecords.Text := NumberRecords.ToString;
//
// RecordSlider.SliderMin := 0.0;
// RecordSlider.SliderMax := MdfNumberRecords;
// RecordSlider.SliderPosition := StartRecord;
// RecordSlider.SliderSpan := NumberRecords;
// SliderSpan := NumberRecords;
//
// RecordSlider.Enabled := True;
// EdStartRecord.Enabled := True;
// EdNumberRecords.Enabled := True;
// BtnStart.Enabled := True;
// BtnEnd.Enabled := True;
// BtnPrev.Enabled := True;
// BtnNext.Enabled := True;
// BtnReadData.Enabled := True;
// BtnUnDoZoom.Enabled := True;
// ChkCursor.Enabled := True;
// ChkMeasDate.Enabled := True;
// ChkReadOnSlide.Enabled := True;
//
// AnnoTool.Active := False;
//
// Self.BtnReadDataClick(Self);
// ViewChart.Invalidate;


// ---
// IdxY := CursorTool.NearestPoint(cssVertical, Delta);
//
// // Cursor position on X scale
// // does not work: screenpos and x scale
// // XPosValue := ViewChart.BottomAxis.CalcXPosValue(X);   // -32k ...?!
// // XPosValue := ViewChart.BottomAxis.CalcPosValue(X);    // -32k ...?!
// // works: screesnpos and series
// // Series.XScreenToValue(X)
//
// // cursor all signals
// S := '';
// case BottomAxisTimeType of
// xtMeas:
// begin
// StrUnit := '[s]';
// S := S + 'X: ' + FloatToStrF(Series.XScreenToValue(X), ffFixed, 10, 3);
// S := S + ' ' + StrUnit;
// HtmlColorDecorator(S, clBlack);
// for var Idx := 0 to NumSignals - 1 do
// begin
// if Signals[Idx].Serie.Visible then
// begin
// S := S + #13;
// if Signals[Idx].Serie = CursorTool.Series then
// S := S + '*'
// else
// S := S + ' ';
// StrName := Signals[Idx].Serie.Title + ': ';
// StrNumber := FloatToStrF(Signals[Idx].Serie.YValue[IdxY], ffFixed, 10, 3);
// StrNumber := Format('%10s', [StrNumber]);
// S := S + HtmlColorDecorator(StrName + StrNumber, Signals[Idx].Serie.Color);
// end;
// end;
// end;
// xtDate:
// begin
// Date := Series.XScreenToValue(X);
// S := S + 'X: ' + Date.Format('hh:mm:ss') + ',' + Date.GetMilliSecond.ToString;
// HtmlColorDecorator(S, clBlack);
// for var Idx := 0 to NumSignals - 1 do
// begin
// if Signals[Idx].Serie.Visible then
// begin
// S := S + #13;
// if Signals[Idx].Serie = CursorTool.Series then
// S := S + '*'
// else
// S := S + ' ';
// StrName := Signals[Idx].Serie.Title + ': ';
// StrNumber := FloatToStrF(Signals[Idx].Serie.YValue[IdxY], ffFixed, 10, 3);
// StrNumber := Format('%15s', [StrNumber]);
// S := S + HtmlColorDecorator(StrName + StrNumber, Signals[Idx].Serie.Color);
// end;
// end;
// end;
// end;
//
// RectTool.Shape.Text := S;
// // ---
// // // Cursor: single signal
// // S := '';
// // case BottomAxisTimeType of
// // xtMeas:
// // begin
// // S := S + 'Signal: ' + CursorTool.Series.Title + #13;
// // S := S + 'X: ' + FloatToStrF(Series.XScreenToValue(X), ffFixed, 10, 3);
// // // S := S + ' Y: ' + FloatToStrF(Series.YValue[IdxY], ffFixed, 10, 3);
// // S := S + ' / ' + HtmlColorDecorator(' Y: ' + FloatToStrF(Series.YValue[IdxY], ffFixed, 10, 3), CursorTool.Series.Color);
// // end;
// // xtDate:
// // begin
// // Date := Series.XScreenToValue(X);
// // // S := S + Date.Format('ddddd') + #13;
// // S := S + 'Signal: ' + CursorTool.Series.Title + #13;
// // S := S + 'X: ' + Date.Format('hh:mm:ss') + ',' + Date.GetMilliSecond.ToString;
// // // S := S + ' / Y: ' + FloatToStrF(Series.YValue[IdxY], ffFixed, 10, 3);
// // S := S + ' / ' + HtmlColorDecorator('Y: ' + FloatToStrF(Series.YValue[IdxY], ffFixed, 10, 3), CursorTool.Series.Color);
// // end;
// // end;


// ---
// w:= Chart1.Canvas.TextWidth(floattostr(Round(Chart1.Axes.Left.Maximum)));
// Chart1.Axes.Left.LabelsSize := w;
//
// ViewChart.LeftAxis.Grid.Color := GridColor;
// ViewChart.LeftAxis.Title.Text := '';
// ViewChart.LeftAxis.MinorTickCount := 1;
// ViewChart.LeftAxis.Increment := 2.5;
// ViewChart.LeftAxis.Title.Position := tpEnd;
// ViewChart.LeftAxis.Title.Alignment := taRightJustify;

// funktioniert nicht
//
// ViewChart.Walls.Visible := False;   // komplett zeigen/verstecken ist das einzige was wirkung zeigt
// ViewChart.View3DWalls := False;     // dito
//
// ViewChart.Walls.Left.Bevel := bvNone;
// ViewChart.Walls.Left.Color := ScaleColor;
// ViewChart.Walls.Left.Size := 0;
// ViewChart.Walls.Left.Visible := True; // True/False egal, immer vorhanden
// ViewChart.Walls.Bottom.Visible := False;
// ViewChart.Walls.Back.Visible := False; // zeigt GAR keine Rnder mehr auch wenn Left.Visible = True
//
// ViewChart.Walls.Bottom.Color := ScaleColor;
// ViewChart.Walls.Right.Color := ScaleColor;
// ViewChart.Walls.Left.Pen.Color := ScaleColor;
// ViewChart.Walls.Visible := False;   // komplett verstecken tut's, aber dann kein oberer/rechter Rand mehr
// ViewChart.Walls.Size := 1; // 5; // 50;
// ViewChart.Walls.Left.Visible := False;
// ViewChart.Walls.Right.Visible := True;
// ViewChart.Walls.Right.Color := ScaleColor;
// ViewChart.Walls.Right.Pen.Color := ScaleColor;
//
// ViewChart.Border.Color := ScaleColor;
// ViewChart.Border.SmallDots := True;
// ViewChart.Border.Visible := True;
// ViewChart.BackColor := clTeeColor; //clGray;
//
// ViewChart.Canvas.Pen.Color := clRed; //clBlack;
// ViewChart.Canvas.Pen.Width := 1;
// ViewChart.Canvas.Pen.Style := psDot;
// ViewChart.Canvas.Brush.Color := clWhite; // clGray // does not work
// ViewChart.Canvas.Brush.Style := bsSolid;


// test ((
// var
// Width, Offset, TitleWidth, TickLength, test: Integer;
// Offset := 0;
// for var Idx := 0 to NumSignals - 1 do
// begin
// if Signals[Idx].Serie.Visible then
// begin
// Scales[Idx].Shape.Visible := True;
// Scales[Idx].Shape.Color := clBlue;
// Scales[Idx].Shape.Width := 1;
// // Scales[Idx].LabelsSize := 0; //Scales[Idx].LabelsSize div 2;
// test := Scales[Idx].LabelsSize;
// Scales[Idx].LabelsSize := 0;
//
// Width := Scales[Idx].MaxLabelsWidth;
// // ScaleWidth := Scales[idx].Shape.Width;
// TitleWidth := Scales[Idx].Title.ShapeBounds.Width;
// TickLength := Scales[Idx].TickLength;
// // TitleWidth := Round(Scales[Idx].Title.Font.SizeFloat); // Trunc
// // Scales[Idx].LabelsSize := -5;
// // Scales[Idx].PosLabels := ... no, read-only
// Scales[Idx].PositionUnits := muPixels;
// ScaleWidth := (Width + TickLength + TitleWidth);
// Scales[Idx].PositionPercent := ScaleWidth * Offset; // auch bei Pixelbezug 'Position*Percent*'!
// Inc(Offset);
// end;
// end;

// test ))

// funktioniert nicht
//
// Skalen links von chart re-positionieren (fontgre usw beachten)
// var
// Width, Offset, TitleWidth, TickLength: Integer;
// Offset := 0;
// for var Idx := 0 to NumSignals - 1 do
// begin
// if Signals[Idx].Serie.Visible then
// begin
// Width := Scales[Idx].MaxLabelsWidth;
// // ScaleWidth := Scales[idx].Shape.Width;
// TitleWidth := Scales[Idx].Title.ShapeBounds.Width;
// TickLength := Scales[Idx].TickLength;
// // TitleWidth := Round(Scales[Idx].Title.Font.SizeFloat); // Trunc
// // Scales[Idx].LabelsSize := -5;
// // Scales[Idx].PosLabels := ... no, read-only
// Scales[Idx].PositionUnits := muPixels;
// ScaleWidth := (Width + TickLength + TitleWidth);
// Scales[Idx].PositionPercent := ScaleWidth * Offset; // auch bei Pixelbezug 'Position*Percent*'!
// Inc(Offset);
// end;
// end;
// test ))


// funktioniert nicht
//
// Offset := 0;
// for var Idx := 0 to NumSignals - 1 do
// begin
// if Signals[Idx].Serie.Visible then
// begin
// // Scales[Idx].PositionPercent := 3.0 * Offset;
// Scales[Idx].PositionPercent := ScaleWidth * Offset;
// Inc(Offset);
// end;
// end;
// ViewChart.BottomAxis.StartPosition := 3.0 * (Offset - 1);
// ViewChart.BottomAxis.StartPosition := (ScaleWidth * (Offset - 1)) / ViewChart.Width * 100;
// ViewChart.BottomAxis.PositionUnits := muPixels;
// ViewChart.BottomAxis.StartPosition := ScaleWidth * (Offset - 1);

// funktioniert auch nicht
//
// var
// Width, TitleWidth, TickLength: Integer;
// Offset := 0;
// for var Idx := 0 to NumSignals - 1 do
// begin
// if Signals[Idx].Serie.Visible then
// begin
// Width := Scales[Idx].MaxLabelsWidth;
// // ScaleWidth := Scales[idx].Shape.Width;
// TitleWidth := Scales[Idx].Title.ShapeBounds.Width;
// TickLength := Scales[Idx].TickLength;
// // TitleWidth := Round(Scales[Idx].Title.Font.SizeFloat); // Trunc
// // Scales[Idx].LabelsSize := -5;
// // Scales[Idx].PosLabels := ... no, read-only
// Scales[Idx].PositionUnits := muPixels;
// ScaleWidth := (Width + TickLength + TitleWidth);
// Scales[Idx].PositionPercent := ScaleWidth * Offset; // auch bei Pixelbezug 'Position*Percent*'!
// Inc(Offset);
// end;
// end;
// ViewChart.BottomAxis.PositionUnits := muPixels;
// ViewChart.BottomAxis.StartPosition := ScaleWidth * (Offset - 1);

// ---
// case BottomAxisTimeType of
// xtMeas:
// begin
// for var SignalIdx := 0 to NumSignals - 1 do
// begin
// pSerie := @Signals[SignalIdx].Serie;
// pSerie.BeginUpdate;
// pSerie.Clear;
//
// pValue := @ByteBuffer[Signals[SignalIdx].Offset];
// // pByte := @ByteBuffer[Signals[SignalIdx].Offset];
// for var SampleIdx := 0 to NumberRecords - 1 do
// begin
// TimeStamp := (StartRecord + SampleIdx) * pChanGrp.SampleDistance;
// pSerie.AddXY(TimeStamp, pValue^);
// Inc(pValue, pChanGrp.CNNumber - 1); // todo, +Recordsize -> pValue = Byte-Pointer
// pByte := @pValue;
// // Inc(pByte, pChanGrp.RecordSize);
// // Inc(pByte, Signals[SignalIdx].Bytes * (pChanGrp.CNNumber - 1)); // all values of same size
// end;
// pSerie.EndUpdate;
// pSerie.ParentChart := ViewChart; // nil: remove from chart
// end;
// end;
// xtDate:
// begin
// for var SignalIdx := 0 to NumSignals - 1 do
// begin
// pSerie := @Signals[SignalIdx].Serie;
// pSerie.BeginUpdate;
// pSerie.Clear;
//
// pValue := @ByteBuffer[Signals[SignalIdx].Offset];
// for var SampleIdx := 0 to NumberRecords - 1 do
// begin
// Date := IncMilliSecond(MdfFileStartTime, round((StartRecord + SampleIdx) * pChanGrp.SampleDistance * 1000));
// pSerie.AddXY(Date, pValue^);
// Inc(pValue, pChanGrp.CNNumber - 1);
// end;
// pSerie.EndUpdate;
// pSerie.ParentChart := ViewChart;
// end;
// end;
// end;
